"""Prepare static features from static + dynamic features
"""
import os
from concurrent.futures import ProcessPoolExecutor
from os.path import join

import hydra
import numpy as np
import pyworld
from hydra.utils import to_absolute_path
from nnsvs.logger import getLogger
from nnsvs.multistream import get_static_features
from nnsvs.util import get_world_stream_info, load_utt_list
from omegaconf import DictConfig, OmegaConf
from tqdm import tqdm


def _extract_static_features(
    in_dir,
    out_dir,
    utt_id,
    num_windows,
    stream_sizes,
    has_dynamic_features,
    mgc2sp=False,
    sample_rate=48000,
) -> None:
    feats = np.load(join(in_dir, utt_id + "-feats.npy"))

    assert np.any(has_dynamic_features)
    streams = get_static_features(
        feats.reshape(1, -1, feats.shape[-1]),
        num_windows,
        stream_sizes,
        has_dynamic_features,
    )

    # remove batch-axis
    streams = list(map(lambda x: x.squeeze(0), streams))

    # Convert mgc2sp
    if mgc2sp:
        mgc = streams[0]
        fft_size = pyworld.get_cheaptrick_fft_size(sample_rate)
        sp = np.log(
            pyworld.decode_spectral_envelope(
                mgc.astype(np.float64), sample_rate, fft_size
            ).astype(np.float32)
        )
        streams[0] = sp

    static_feats = np.concatenate(streams, axis=-1).astype(np.float32)

    static_path = join(out_dir, utt_id + "-feats.npy")
    np.save(static_path, static_feats, allow_pickle=False)


@hydra.main(config_path="conf/prepare_static_features", config_name="config")
def my_app(config: DictConfig) -> None:
    logger = getLogger(config.verbose)
    logger.info(OmegaConf.to_yaml(config))

    utt_list = to_absolute_path(config.utt_list)
    in_dir = to_absolute_path(config.in_dir)
    out_dir = to_absolute_path(config.out_dir)

    utt_ids = load_utt_list(utt_list)

    stream_sizes, has_dynamic_features = get_world_stream_info(
        config.acoustic.sample_rate,
        config.acoustic.mgc_order,
        config.acoustic.num_windows,
        config.acoustic.vibrato_mode,
    )

    os.makedirs(out_dir, exist_ok=True)

    with ProcessPoolExecutor(max_workers=config.max_workers) as executor:
        futures = [
            executor.submit(
                _extract_static_features,
                in_dir,
                out_dir,
                utt_id,
                config.acoustic.num_windows,
                stream_sizes,
                has_dynamic_features,
                mgc2sp=config.mgc2sp,
                sample_rate=config.sample_rate,
            )
            for utt_id in utt_ids
        ]
        for future in tqdm(futures):
            future.result()


def entry():
    my_app()  # pylint: disable=no-value-for-parameter


if __name__ == "__main__":
    my_app()  # pylint: disable=no-value-for-parameter



"""Discriminator implementations mostly used for GAN-based post-filters.

All the discriminators must returns list of tensors.
The last tensor of the list is regarded as the output of the discrminator.
The others are used as intermedieate feature maps.
"""

import numpy as np
import torch
from nnsvs.util import init_weights
from torch import nn


class Conv2dD(nn.Module):
    """Conv2d-based discriminator


    The implementation follows the discrimiantor of the GAN-based post-filters
    in :cite:t:`Kaneko2017Interspeech`.

    Args:
        in_dim (int): Input feature dim
        channels (int): Number of channels
        kernel_size (tuple): Kernel size for 2d-convolution
        padding (tuple): Padding for 2d-convolution
        last_sigmoid (bool): If True, apply sigmoid on the output
        init_type (str): Initialization type
        padding_mode (str): Padding mode
    """

    def __init__(
        self,
        in_dim=None,
        channels=64,
        kernel_size=(5, 3),
        padding=(0, 0),
        last_sigmoid=False,
        init_type="kaiming_normal",
        padding_mode="zeros",
    ):
        super().__init__()
        self.last_sigmoid = last_sigmoid
        C = channels
        ks = np.asarray(list(kernel_size))
        if padding is None:
            padding = (ks - 1) // 2

        self.convs = nn.ModuleList()
        self.convs.append(
            nn.Sequential(
                nn.Conv2d(
                    1,
                    C,
                    kernel_size=ks,
                    padding=padding,
                    stride=(1, 1),
                    padding_mode=padding_mode,
                ),
                nn.LeakyReLU(0.2),
            )
        )
        self.convs.append(
            nn.Sequential(
                nn.Conv2d(
                    C,
                    2 * C,
                    kernel_size=ks,
                    padding=padding,
                    stride=(2, 1),
                    padding_mode=padding_mode,
                ),
                nn.LeakyReLU(0.2),
            )
        )
        self.convs.append(
            nn.Sequential(
                nn.Conv2d(
                    2 * C,
                    4 * C,
                    kernel_size=ks,
                    padding=padding,
                    stride=(2, 1),
                    padding_mode=padding_mode,
                ),
                nn.LeakyReLU(0.2),
            )
        )
        self.convs.append(
            nn.Sequential(
                nn.Conv2d(
                    4 * C,
                    2 * C,
                    kernel_size=ks,
                    padding=padding,
                    stride=(2, 1),
                    padding_mode=padding_mode,
                ),
                nn.LeakyReLU(0.2),
            )
        )
        self.last_conv = nn.Conv2d(
            2 * C,
            1,
            kernel_size=ks,
            padding=padding,
            stride=(1, 1),
            padding_mode=padding_mode,
        )
        init_weights(self, init_type)

    def forward(self, x, c=None, lengths=None):
        """Forward step

        Args:
            x (torch.Tensor): Input tensor
            c (torch.Tensor): Optional conditional features
            lengths (torch.Tensor): Optional lengths of the input

        Returns:
            list: List of output tensors
        """
        outs = []
        # (B, T, C) -> (B, 1, T, C):
        x = x.unsqueeze(1)
        for conv in self.convs:
            x = conv(x)
            outs.append(x)
        y = self.last_conv(x)
        y = torch.sigmoid(y) if self.last_sigmoid else y
        # (B, 1, T, C) -> (B, T, C)
        y = y.squeeze(1)
        outs.append(y)

        return [outs]



"""Prepare input features for training neural vocoders
"""
import os
from concurrent.futures import ProcessPoolExecutor
from os.path import exists, islink, join

import hydra
import numpy as np
from hydra.utils import to_absolute_path
from nnsvs.logger import getLogger
from nnsvs.multistream import get_static_features
from nnsvs.util import get_world_stream_info, load_utt_list
from omegaconf import DictConfig, OmegaConf
from tqdm import tqdm


def _prepare_voc_features(
    in_dir,
    out_dir,
    utt_id,
    num_windows,
    stream_sizes,
    has_dynamic_features,
) -> None:
    feats = np.load(join(in_dir, utt_id + "-feats.npy"))
    in_wave_path = join(in_dir, utt_id + "-wave.npy")
    assert exists(in_wave_path)

    streams = get_static_features(
        feats.reshape(1, -1, feats.shape[-1]),
        num_windows,
        stream_sizes,
        has_dynamic_features,
    )

    # remove batch-axis
    streams = list(map(lambda x: x.squeeze(0), streams))

    # NOTE: even if the number of streams are larger than 4, we only use the first 4 streams
    # for training neural vocoders
    if len(streams) >= 4:
        mgc, lf0, vuv, bap = streams[0], streams[1], streams[2], streams[3]
        voc_feats = np.hstack((mgc, lf0, vuv, bap)).astype(np.float32)
    elif len(streams) == 3:
        mel, lf0, vuv = streams[0], streams[1], streams[2]
        voc_feats = np.hstack((mel, lf0, vuv)).astype(np.float32)

    voc_feats_path = join(out_dir, utt_id + "-feats.npy")
    np.save(voc_feats_path, voc_feats, allow_pickle=False)

    # NOTE: To train vocoders with https://github.com/kan-bayashi/ParallelWaveGAN
    # target waveform needs to be created in the same directory as the vocoder input features.
    save_wave_path = join(out_dir, utt_id + "-wave.npy")
    if (not exists(save_wave_path)) and (not islink(save_wave_path)):
        os.symlink(join(in_dir, utt_id + "-wave.npy"), save_wave_path)


@hydra.main(config_path="conf/prepare_static_features", config_name="config")
def my_app(config: DictConfig) -> None:
    logger = getLogger(config.verbose)
    logger.info(OmegaConf.to_yaml(config))

    utt_list = to_absolute_path(config.utt_list)
    in_dir = to_absolute_path(config.in_dir)
    out_dir = to_absolute_path(config.out_dir)

    utt_ids = load_utt_list(utt_list)

    if config.acoustic.relative_f0:
        raise ValueError("Relative F0 is not supported")

    if config.acoustic.feature_type == "world":
        stream_sizes = get_world_stream_info(
            config.acoustic.sample_rate,
            config.acoustic.mgc_order,
            config.acoustic.num_windows,
            config.acoustic.vibrato_mode,
            use_mcep_aperiodicity=config.acoustic.use_mcep_aperiodicity,
            mcep_aperiodicity_order=config.acoustic.mcep_aperiodicity_order,
        )
    elif config.acoustic.feature_type == "melf0":
        stream_sizes = [config.acoustic.num_mels, 1, 1]
    else:
        raise ValueError(f"Unknown feature type: {config.acoustic.feature_type}")

    os.makedirs(out_dir, exist_ok=True)
    with ProcessPoolExecutor(max_workers=config.max_workers) as executor:
        futures = [
            executor.submit(
                _prepare_voc_features,
                in_dir,
                out_dir,
                utt_id,
                config.acoustic.num_windows,
                stream_sizes,
                config.acoustic.dynamic_features_flags,
            )
            for utt_id in utt_ids
        ]
        for future in tqdm(futures):
            future.result()


def entry():
    my_app()  # pylint: disable=no-value-for-parameter


if __name__ == "__main__":
    my_app()  # pylint: disable=no-value-for-parameter



import torch
from nnsvs.wavenet import conv
from torch import nn


def Conv1d(in_channels, out_channels, kernel_size, *args, **kwargs):
    """Weight-normalized Conv1d layer."""
    m = conv.Conv1d(in_channels, out_channels, kernel_size, *args, **kwargs)
    return nn.utils.weight_norm(m)


def Conv1d1x1(in_channels, out_channels, bias=True):
    """1x1 Weight-normalized Conv1d layer."""
    return Conv1d(in_channels, out_channels, kernel_size=1, bias=bias)


class ResSkipBlock(nn.Module):
    """Convolution block with residual and skip connections.

    Args:
        residual_channels (int): Residual connection channels.
        gate_channels (int): Gated activation channels.
        kernel_size (int): Kernel size of convolution layers.
        skip_out_channels (int): Skip connection channels.
        dilation (int): Dilation factor.
        cin_channels (int): Local conditioning channels.
        args (list): Additional arguments for Conv1d.
        kwargs (dict): Additional arguments for Conv1d.
    """

    def __init__(
        self,
        residual_channels,  # 残差結合のチャネル数
        gate_channels,  # ゲートのチャネル数
        kernel_size,  # カーネルサイズ
        skip_out_channels,  # スキップ結合のチャネル数
        dilation=1,  # dilation factor
        cin_channels=80,  # 条件付特徴量のチャネル数
        *args,
        **kwargs,
    ):
        super().__init__()
        self.padding = (kernel_size - 1) * dilation

        # 1 次元膨張畳み込み (dilation == 1 のときは、通常の1 次元畳み込み)
        self.conv = Conv1d(
            residual_channels,
            gate_channels,
            kernel_size,
            *args,
            padding=self.padding,
            dilation=dilation,
            **kwargs,
        )

        # local conditioning 用の 1x1 convolution
        self.conv1x1c = Conv1d1x1(cin_channels, gate_channels, bias=False)

        # ゲート付き活性化関数のために、1 次元畳み込みの出力は2 分割されることに注意
        gate_out_channels = gate_channels // 2
        self.conv1x1_out = Conv1d1x1(gate_out_channels, residual_channels)
        self.conv1x1_skip = Conv1d1x1(gate_out_channels, skip_out_channels)

    def forward(self, x, c):
        """Forward step

        Args:
            x (torch.Tensor): Input signal.
            c (torch.Tensor): Local conditioning signal.

        Returns:
            tuple: Tuple of output signal and skip connection signal
        """
        return self._forward(x, c, False)

    def incremental_forward(self, x, c):
        """Incremental forward

        Args:
            x (torch.Tensor): Input signal.
            c (torch.Tensor): Local conditioning signal.

        Returns:
            tuple: Tuple of output signal and skip connection signal
        """
        return self._forward(x, c, True)

    def _forward(self, x, c, is_incremental):
        # 残差接続用に入力を保持
        residual = x

        # メインの dilated convolutionの計算
        # 推論時と学習時で入力のテンソルのshapeが異なるのに注意
        if is_incremental:
            splitdim = -1  # (B, T, C)
            x = self.conv.incremental_forward(x)
        else:
            splitdim = 1  # (B, C, T)
            x = self.conv(x)
            # 因果性を保証するために、出力をシフトする
            x = x[:, :, : -self.padding]

        # チャンネル方向で出力を分割
        a, b = x.split(x.size(splitdim) // 2, dim=splitdim)

        # local conditioning
        c = self._conv1x1_forward(self.conv1x1c, c, is_incremental)
        ca, cb = c.split(c.size(splitdim) // 2, dim=splitdim)
        a, b = a + ca, b + cb

        # ゲート付き活性化関数
        x = torch.tanh(a) * torch.sigmoid(b)

        # スキップ接続用の出力を計算
        s = self._conv1x1_forward(self.conv1x1_skip, x, is_incremental)

        # 残差接続の要素和行う前に、次元数を合わせる
        x = self._conv1x1_forward(self.conv1x1_out, x, is_incremental)

        x = x + residual

        return x, s

    def _conv1x1_forward(self, conv, x, is_incremental):
        if is_incremental:
            x = conv.incremental_forward(x)
        else:
            x = conv(x)
        return x

    def clear_buffer(self):
        """Clear input buffer."""
        for c in [
            self.conv,
            self.conv1x1_out,
            self.conv1x1_skip,
            self.conv1x1c,
        ]:
            if c is not None:
                c.clear_buffer()



from os.path import join

import librosa
import numpy as np
import pysptk
import pyworld
import soundfile as sf
from nnmnkwii.datasets import FileDataSource
from nnmnkwii.frontend import merlin as fe
from nnmnkwii.io import hts
from nnmnkwii.preprocessing.f0 import interp1d
from nnmnkwii.util import apply_delta_windows
from nnsvs.io.hts import get_pitch_index, get_pitch_indices
from nnsvs.multistream import get_windows
from nnsvs.pitch import (
    compute_f0_correction_ratio,
    extract_smoothed_continuous_f0,
    extract_smoothed_f0,
    extract_vibrato_likelihood,
    extract_vibrato_parameters,
    hz_to_cent_based_c4,
    lowpass_filter,
)
from nnsvs.util import init_seed
from parallel_wavegan.bin.preprocess import logmelfilterbank
from scipy.signal import firwin, lfilter


def _collect_files(data_root, utt_list, ext):
    with open(utt_list) as f:
        files = f.readlines()
    files = map(lambda utt_id: utt_id.strip(), files)
    files = filter(lambda utt_id: len(utt_id) > 0, files)
    files = list(map(lambda utt_id: join(data_root, f"{utt_id}{ext}"), files))
    return files


def _midi_to_hz(x, idx, log_f0=False):
    z = np.zeros(len(x))
    indices = x[:, idx] > 0
    z[indices] = librosa.midi_to_hz(x[indices, idx])
    if log_f0:
        z[indices] = np.log(z[indices])
    return z


def low_cut_filter(x, fs, cutoff=70):
    """Low cut filter

    Args:
        x (ndarray): Waveform sequence
        fs (int): Sampling frequency
        cutoff (float): Cutoff frequency of low cut filter

    Return:
        (ndarray): Low cut filtered waveform sequence

    """
    nyquist = fs // 2
    norm_cutoff = cutoff / nyquist
    fil = firwin(255, norm_cutoff, pass_zero=False)
    lcf_x = lfilter(fil, 1, x)

    return lcf_x


class MusicalLinguisticSource(FileDataSource):
    def __init__(
        self,
        utt_list,
        data_root,
        question_path,
        add_frame_features=False,
        subphone_features=None,
        log_f0_conditioning=True,
        frame_period=5,
    ):
        self.data_root = data_root
        self.utt_list = utt_list
        self.add_frame_features = add_frame_features
        self.subphone_features = subphone_features
        self.binary_dict, self.numeric_dict = hts.load_question_set(
            question_path, append_hat_for_LL=False
        )
        self.log_f0_conditioning = log_f0_conditioning
        self.frame_period = frame_period
        self.pitch_indices = get_pitch_indices(self.binary_dict, self.numeric_dict)

    def collect_files(self):
        return _collect_files(self.data_root, self.utt_list, ".lab")

    def collect_features(self, path):
        labels = hts.load(path)
        hts_frame_shift = int(self.frame_period * 1e4)
        labels.frame_shift = hts_frame_shift
        features = fe.linguistic_features(
            labels,
            self.binary_dict,
            self.numeric_dict,
            add_frame_features=self.add_frame_features,
            subphone_features=self.subphone_features,
            frame_shift=hts_frame_shift,
        )
        if self.log_f0_conditioning:
            for idx in self.pitch_indices:
                lf0_score = _midi_to_hz(features, idx, True)
                features[:, idx] = interp1d(lf0_score, kind="slinear")

        return features.astype(np.float32)


class TimeLagFeatureSource(FileDataSource):
    def __init__(self, utt_list, label_phone_score_dir, label_phone_align_dir):
        self.utt_list = utt_list
        self.label_phone_score_dir = label_phone_score_dir
        self.label_phone_align_dir = label_phone_align_dir

    def collect_files(self):
        labels_score = _collect_files(self.label_phone_score_dir, self.utt_list, ".lab")
        labels_align = _collect_files(self.label_phone_align_dir, self.utt_list, ".lab")
        return labels_score, labels_align

    def collect_features(self, label_score_path, label_align_path):
        label_score = hts.load(label_score_path)
        label_align = hts.load(label_align_path)
        timelag = np.asarray(label_align.start_times) - np.asarray(
            label_score.start_times
        )
        # 100ns -> num frames
        timelag = timelag.astype(np.float32) / 50000
        return timelag.reshape(-1, 1)


class DurationFeatureSource(FileDataSource):
    def __init__(self, utt_list, data_root):
        self.utt_list = utt_list
        self.data_root = data_root

    def collect_files(self):
        return _collect_files(self.data_root, self.utt_list, ".lab")

    def collect_features(self, path):
        labels = hts.load(path)
        features = fe.duration_features(labels)
        return features.astype(np.float32)


class WORLDAcousticSource(FileDataSource):
    def __init__(
        self,
        utt_list,
        wav_root,
        label_root,
        question_path,
        f0_extractor="harvest",
        f0_floor=150,
        f0_ceil=700,
        frame_period=5,
        mgc_order=59,
        num_windows=3,
        relative_f0=True,
        interp_unvoiced_aperiodicity=True,
        vibrato_mode="none",  # diff, sine
        sample_rate=48000,
        d4c_threshold=0.85,
        trajectory_smoothing=False,
        trajectory_smoothing_cutoff=50,
        trajectory_smoothing_f0=True,
        trajectory_smoothing_cutoff_f0=20,
        correct_vuv=False,
        correct_f0=False,
        dynamic_features_flags=None,
        use_world_codec=False,
        use_mcep_aperiodicity=False,
        mcep_aperiodicity_order=24,
        res_type="scipy",
    ):
        self.utt_list = utt_list
        self.wav_root = wav_root
        self.label_root = label_root
        self.binary_dict, self.numeric_dict = hts.load_question_set(
            question_path, append_hat_for_LL=False
        )
        self.pitch_idx = get_pitch_index(self.binary_dict, self.numeric_dict)
        self.f0_extractor = f0_extractor
        self.f0_floor = f0_floor
        self.f0_ceil = f0_ceil
        self.frame_period = frame_period
        self.mgc_order = mgc_order
        self.relative_f0 = relative_f0
        self.interp_unvoiced_aperiodicity = interp_unvoiced_aperiodicity
        self.vibrato_mode = vibrato_mode
        self.windows = get_windows(num_windows)
        self.sample_rate = sample_rate
        self.d4c_threshold = d4c_threshold
        self.trajectory_smoothing = trajectory_smoothing
        self.trajectory_smoothing_cutoff = trajectory_smoothing_cutoff
        self.trajectory_smoothing_f0 = trajectory_smoothing_f0
        self.trajectory_smoothing_cutoff_f0 = trajectory_smoothing_cutoff_f0
        self.correct_vuv = correct_vuv
        self.correct_f0 = correct_f0
        self.use_world_codec = use_world_codec
        self.use_mcep_aperiodicity = use_mcep_aperiodicity
        self.mcep_aperiodicity_order = mcep_aperiodicity_order
        if dynamic_features_flags is None:
            # NOTE: we have up to 6 streams: (mgc, lf0, vuv, bap, vib, vib_flags)
            dynamic_features_flags = [True, True, False, True, True, False]
        self.dynamic_features_flags = dynamic_features_flags
        self.res_type = res_type

    def collect_files(self):
        wav_paths = _collect_files(self.wav_root, self.utt_list, ".wav")
        label_paths = _collect_files(self.label_root, self.utt_list, ".lab")
        return wav_paths, label_paths

    def collect_features(self, wav_path, label_path):
        labels = hts.load(label_path)
        hts_frame_shift = int(self.frame_period * 1e4)
        labels.frame_shift = hts_frame_shift
        num_frames = int(labels.num_frames(frame_shift=hts_frame_shift))

        l_features = fe.linguistic_features(
            labels,
            self.binary_dict,
            self.numeric_dict,
            add_frame_features=True,
            subphone_features="coarse_coding",
            frame_shift=hts_frame_shift,
        )

        f0_score = _midi_to_hz(l_features, self.pitch_idx, False)
        notes = l_features[:, self.pitch_idx]
        notes = notes[notes > 0]

        # allow 200 cent upper and 600 cent lower to properly handle F0 estimation of
        # preparation, vibrato and overshoot.
        # NOET: set the minimum f0 to 63.5 Hz (125 - 3*20.5)
        # https://acoustics.jp/qanda/answer/50.html
        # NOTE: sinsy allows 30-150 cent frequency range for vibrato (as of 2010)
        # https://staff.aist.go.jp/m.goto/PAPER/SIGMUS201007oura.pdf
        min_f0 = max(63.5, librosa.midi_to_hz(min(notes) - 6))
        max_f0 = librosa.midi_to_hz(max(notes) + 2)
        assert max_f0 > min_f0

        # Use fixed f0 range for HARVEST if specified
        if self.f0_floor is not None:
            min_f0 = self.f0_floor
        if self.f0_ceil is not None:
            max_f0 = self.f0_ceil

        # Workaround segfault issues of WORLD's CheapTrick
        min_f0 = min(min_f0, 500)

        x, fs = sf.read(wav_path)
        assert np.max(x) <= 1.0
        assert x.dtype == np.float64

        if fs != self.sample_rate:
            x = librosa.resample(
                x, orig_sr=fs, target_sr=self.sample_rate, res_type=self.res_type
            )
            fs = self.sample_rate

        if self.f0_extractor == "parselmouth":
            import parselmouth

            assert (
                self.f0_floor is not None and self.f0_ceil is not None
            ), "must be set manually"
            harvest_num_frames = int(int(1000 * len(x) / fs) / self.frame_period) + 1
            f0 = (
                parselmouth.Sound(x.astype(np.float64), fs)
                .to_pitch_ac(
                    time_step=self.frame_period * 0.001,
                    voicing_threshold=0.6,
                    very_accurate=False,
                    pitch_floor=min_f0,
                    pitch_ceiling=max_f0,
                )
                .selected_array["frequency"]
            )
            pad = int(np.round((3 / min_f0) / (self.frame_period * 0.001)))
            f0 = np.pad(f0, [[0, pad]], mode="constant")
            if len(f0) > harvest_num_frames:
                f0 = f0[:harvest_num_frames]
            elif len(f0) < harvest_num_frames:
                f0 = np.pad(f0, (0, harvest_num_frames - len(f0)), mode="constant")

            timeaxis = np.arange(harvest_num_frames) * self.frame_period * 0.001
        elif self.f0_extractor == "harvest":
            f0, timeaxis = pyworld.harvest(
                x, fs, frame_period=self.frame_period, f0_floor=min_f0, f0_ceil=max_f0
            )
        elif self.f0_extractor == "dio":
            f0, timeaxis = pyworld.dio(
                x, fs, frame_period=self.frame_period, f0_floor=min_f0, f0_ceil=max_f0
            )
            f0 = pyworld.stonemask(x, f0, timeaxis, fs)
        else:
            raise ValueError(f"unknown f0 extractor: {self.f0_extractor}")

        # Workaround for https://github.com/r9y9/nnsvs/issues/7
        f0 = np.maximum(f0, 0)

        # Correct V/UV (and F0) based on the musical score information
        # treat frames where musical notes are not assigned as unvoiced
        # Use smoothed mask so that we don't mask out overshoot or something
        # that could happen at the start/end of notes
        # 0.5 sec. window (could be tuned for better results)
        if self.correct_vuv:
            win_length = int(0.5 / (self.frame_period * 0.001))
            mask = np.convolve(f0_score, np.ones(win_length) / win_length, "same")
            if len(f0) > len(mask):
                mask = np.pad(mask, (0, len(f0) - len(mask)), "constant")
            elif len(f0) < len(mask):
                mask = mask[: len(f0)]
            f0 = f0 * np.sign(mask)

        spectrogram = pyworld.cheaptrick(x, f0, timeaxis, fs)
        aperiodicity = pyworld.d4c(x, f0, timeaxis, fs, threshold=self.d4c_threshold)

        if np.isnan(aperiodicity).any():
            print(wav_path)
            print(min_f0, max_f0, aperiodicity.shape, fs)
            print(np.isnan(aperiodicity).sum())
            print(aperiodicity)
            raise RuntimeError("Aperiodicity has NaN")

        # Apply pitch correction
        # NOTE: we should better to apply pitch corrections
        # or manually adjust UST/musicxml in advance.
        sr_f0 = int(1 / (self.frame_period * 0.001))
        if self.correct_f0:
            f0_smooth = extract_smoothed_f0(f0, sr_f0, cutoff=20)
            ratio = compute_f0_correction_ratio(f0_smooth, f0_score)
            f0 *= ratio

        lf0 = f0[:, np.newaxis].copy()
        nonzero_indices = np.nonzero(lf0)
        lf0[nonzero_indices] = np.log(f0[:, np.newaxis][nonzero_indices])
        if self.f0_extractor == "harvest":
            # https://github.com/mmorise/World/issues/35#issuecomment-306521887
            vuv = (aperiodicity[:, 0] < 0.5).astype(np.float32)[:, None]
        else:
            vuv = (lf0 != 0).astype(np.float32)

        # F0 -> continuous F0
        lf0 = interp1d(lf0, kind="slinear")

        # Smooth continuous F0 to avoid discontinuities
        if self.trajectory_smoothing_f0:
            lf0 = extract_smoothed_continuous_f0(
                lf0, sr_f0, cutoff=self.trajectory_smoothing_cutoff_f0
            )

        # Fill continuous F0s for segments where no notes are assigned & no F0s are detected.
        lf0_score = _midi_to_hz(l_features, self.pitch_idx, True)
        clf0_score = interp1d(lf0_score, kind="slinear")
        mask = np.convolve(lf0_score, np.ones(1), "same")
        if len(f0) > len(mask):
            mask = np.pad(mask, (0, len(f0) - len(mask)), "constant")
            clf0_score = np.pad(clf0_score, (0, len(f0) - len(clf0_score)), "constant")
        elif len(f0) < len(mask):
            mask = mask[: len(f0)]
            clf0_score = clf0_score[: len(f0)]
        ind = (mask + f0.reshape(-1)) <= 0
        lf0[ind, 0] = clf0_score[ind]

        # Vibrato parameter extraction
        if self.vibrato_mode == "sine":
            win_length = 64
            n_fft = 256
            threshold = 0.12

            if self.f0_extractor == "harvest":
                # NOTE: harvest is not supported here since the current implemented algorithm
                # relies on v/uv flags to find vibrato sections.
                # We use DIO since it provides more accurate v/uv detection in my experience.
                _f0, _timeaxis = pyworld.dio(
                    x,
                    fs,
                    frame_period=self.frame_period,
                    f0_floor=min_f0,
                    f0_ceil=max_f0,
                )
                _f0 = pyworld.stonemask(x, _f0, _timeaxis, fs)
                f0_smooth = extract_smoothed_f0(_f0, sr_f0, cutoff=8)
            else:
                f0_smooth = extract_smoothed_f0(f0, sr_f0, cutoff=8)

            f0_smooth_cent = hz_to_cent_based_c4(f0_smooth)
            vibrato_likelihood = extract_vibrato_likelihood(
                f0_smooth_cent, sr_f0, win_length=win_length, n_fft=n_fft
            )
            vib_flags, m_a, m_f = extract_vibrato_parameters(
                f0_smooth_cent, vibrato_likelihood, sr_f0, threshold=threshold
            )
            m_a = interp1d(m_a, kind="linear")
            m_f = interp1d(m_f, kind="linear")
            vib = np.stack([m_a, m_f], axis=1)
            vib_flags = vib_flags[:, np.newaxis]
        elif self.vibrato_mode == "diff":
            # NOTE: vibrato is known to have 3 ~ 8 Hz range (in general)
            # remove higher frequency than 3 to separate vibrato from the original F0
            f0_smooth = extract_smoothed_f0(f0, sr_f0, cutoff=3)
            assert len(f0.shape) == 1 and len(f0_smooth.shape) == 1
            vib = (f0 - f0_smooth)[:, np.newaxis]
            vib_flags = None
        elif self.vibrato_mode == "none":
            vib, vib_flags = None, None
        else:
            raise RuntimeError("Unknown vibrato mode: {}".format(self.vibrato_mode))

        if self.use_world_codec:
            mgc = pyworld.code_spectral_envelope(spectrogram, fs, self.mgc_order + 1)
        else:
            mgc = pysptk.sp2mc(
                spectrogram, order=self.mgc_order, alpha=pysptk.util.mcepalpha(fs)
            )
        # NOTE: used as the target for post-filters
        spectrogram = np.log(spectrogram)

        # Post-processing for aperiodicy
        # ref: https://github.com/MTG/WGANSing/blob/mtg/vocoder.py
        if self.interp_unvoiced_aperiodicity:
            is_voiced = (vuv > 0).reshape(-1)
            if not np.any(is_voiced):
                pass  # all unvoiced, do nothing
            else:
                for k in range(aperiodicity.shape[1]):
                    aperiodicity[~is_voiced, k] = np.interp(
                        np.where(~is_voiced)[0],
                        np.where(is_voiced)[0],
                        aperiodicity[is_voiced, k],
                    )

        if self.use_mcep_aperiodicity:
            bap = pysptk.sp2mc(
                aperiodicity,
                order=self.mcep_aperiodicity_order,
                alpha=pysptk.util.mcepalpha(fs),
            )
        else:
            bap = pyworld.code_aperiodicity(aperiodicity, fs)

        # Parameter trajectory smoothing
        if self.trajectory_smoothing:
            modfs = int(1 / (self.frame_period * 0.001))
            for d in range(mgc.shape[1]):
                mgc[:, d] = lowpass_filter(
                    mgc[:, d], modfs, cutoff=self.trajectory_smoothing_cutoff
                )
            for d in range(bap.shape[1]):
                bap[:, d] = lowpass_filter(
                    bap[:, d], modfs, cutoff=self.trajectory_smoothing_cutoff
                )

        # Adjust lengths
        sp = spectrogram[:num_frames]
        mgc = mgc[:num_frames]
        lf0 = lf0[:num_frames]
        vuv = vuv[:num_frames]
        bap = bap[:num_frames]
        vib = vib[:num_frames] if vib is not None else None
        vib_flags = vib_flags[:num_frames] if vib_flags is not None else None

        if self.relative_f0:
            # # F0 derived from the musical score
            f0_score = f0_score[:, None]
            if len(f0_score) > len(f0):
                print(
                    "Warning! likely to have mistakes in alignment in {}".format(
                        label_path
                    )
                )
                print(f0_score.shape, f0.shape)
                f0_score = f0_score[: len(f0)]

            lf0_score = f0_score.copy()
            nonzero_indices = np.nonzero(f0_score)
            lf0_score[nonzero_indices] = np.log(f0_score[nonzero_indices])
            lf0_score = interp1d(lf0_score, kind="slinear")
            # relative f0
            diff_lf0 = lf0 - lf0_score
            diff_lf0 = np.clip(diff_lf0, np.log(0.5), np.log(2.0))

            f0_target = diff_lf0
        else:
            f0_target = lf0

        # Compute delta features if necessary
        if self.dynamic_features_flags[0]:
            mgc = apply_delta_windows(mgc, self.windows)
        if self.dynamic_features_flags[1]:
            f0_target = apply_delta_windows(f0_target, self.windows)
        if self.dynamic_features_flags[3]:
            bap = apply_delta_windows(bap, self.windows)
        if vib is not None and self.dynamic_features_flags[4]:
            vib = apply_delta_windows(vib, self.windows)

        # Concat features
        if vib is None and vib_flags is None:
            features = np.hstack((mgc, f0_target, vuv, bap)).astype(np.float32)
            pf_features = np.hstack((sp, f0_target, vuv, bap)).astype(np.float32)
        elif vib is not None and vib_flags is None:
            features = np.hstack((mgc, f0_target, vuv, bap, vib)).astype(np.float32)
            pf_features = np.hstack((sp, f0_target, vuv, bap, vib)).astype(np.float32)
        elif vib is not None and vib_flags is not None:
            features = np.hstack((mgc, f0_target, vuv, bap, vib, vib_flags)).astype(
                np.float32
            )
            pf_features = np.hstack((sp, f0_target, vuv, bap, vib, vib_flags)).astype(
                np.float32
            )
        else:
            raise RuntimeError("Unknown combination of features")

        if len(features) < num_frames:
            print(f"WARN: length mismatch for {wav_path}")
            print(len(features), num_frames)
            return None, None, None

        # Align waveform and features
        wave = x.astype(np.float32)

        # NOTE: since neural vocoders need to perform integer-valued up-sampling
        # (e.g., 120x upsampling with 5ms and 24kHz sampling), we must ensure
        # that the length of the waveform is a multiple of the integer-valued
        # up-sampling factor.
        frame_shift_int = int(fs * self.frame_period / 1000)
        T = int(features.shape[0] * frame_shift_int)
        if len(wave) < T:
            if T - len(wave) > int(fs * (self.frame_period * 0.001)):
                print("Length mismatch", T, len(wave), T - len(wave))
                print(f"Unaligned data: {wav_path} and {label_path}")
                raise RuntimeError("Unaligned data")
            else:
                pass
            wave = np.pad(wave, (0, T - len(wave)))
        assert wave.shape[0] >= T
        wave = wave[:T]

        assert np.isfinite(features).all()
        assert np.isfinite(wave).all()
        assert np.isfinite(pf_features).all()

        return features, wave, pf_features


class MelF0AcousticSource(FileDataSource):
    def __init__(
        self,
        utt_list,
        wav_root,
        label_root,
        question_path,
        f0_extractor="harvest",
        f0_floor=150,
        f0_ceil=700,
        frame_period=5,
        vibrato_mode="none",  # diff, sine
        sample_rate=48000,
        d4c_threshold=0.85,
        trajectory_smoothing_f0=True,
        trajectory_smoothing_cutoff_f0=20,
        correct_vuv=False,
        correct_f0=False,
        res_type="scipy",
        fft_size=512,
        win_length=480,
        hop_size=120,
        fmin=30,
        fmax=None,
        eps=1e-10,
        num_mels=80,
    ):
        self.utt_list = utt_list
        self.wav_root = wav_root
        self.label_root = label_root
        self.binary_dict, self.numeric_dict = hts.load_question_set(
            question_path, append_hat_for_LL=False
        )
        self.pitch_idx = get_pitch_index(self.binary_dict, self.numeric_dict)
        self.f0_extractor = f0_extractor
        self.f0_floor = f0_floor
        self.f0_ceil = f0_ceil
        self.frame_period = frame_period
        self.vibrato_mode = vibrato_mode
        self.sample_rate = sample_rate
        self.d4c_threshold = d4c_threshold
        self.trajectory_smoothing_f0 = trajectory_smoothing_f0
        self.trajectory_smoothing_cutoff_f0 = trajectory_smoothing_cutoff_f0
        self.correct_vuv = correct_vuv
        self.correct_f0 = correct_f0
        self.res_type = res_type
        self.fft_size = fft_size
        self.win_length = win_length
        self.hop_size = hop_size
        self.fmin = fmin
        if fmax is None:
            fmax = sample_rate // 2
        self.fmax = fmax
        self.eps = eps
        self.num_mels = num_mels

    def collect_files(self):
        wav_paths = _collect_files(self.wav_root, self.utt_list, ".wav")
        label_paths = _collect_files(self.label_root, self.utt_list, ".lab")
        return wav_paths, label_paths

    def collect_features(self, wav_path, label_path):
        labels = hts.load(label_path)
        hts_frame_shift = int(self.frame_period * 1e4)
        labels.frame_shift = hts_frame_shift
        num_frames = int(labels.num_frames(frame_shift=hts_frame_shift))

        l_features = fe.linguistic_features(
            labels,
            self.binary_dict,
            self.numeric_dict,
            add_frame_features=True,
            subphone_features="coarse_coding",
            frame_shift=hts_frame_shift,
        )

        f0_score = _midi_to_hz(l_features, self.pitch_idx, False)
        notes = l_features[:, self.pitch_idx]
        notes = notes[notes > 0]

        # allow 200 cent upper and 600 cent lower to properly handle F0 estimation of
        # preparation, vibrato and overshoot.
        # NOET: set the minimum f0 to 63.5 Hz (125 - 3*20.5)
        # https://acoustics.jp/qanda/answer/50.html
        # NOTE: sinsy allows 30-150 cent frequency range for vibrato (as of 2010)
        # https://staff.aist.go.jp/m.goto/PAPER/SIGMUS201007oura.pdf
        min_f0 = max(63.5, librosa.midi_to_hz(min(notes) - 6))
        max_f0 = librosa.midi_to_hz(max(notes) + 2)
        assert max_f0 > min_f0

        # Use fixed f0 range for HARVEST if specified
        if self.f0_floor is not None:
            min_f0 = self.f0_floor
        if self.f0_ceil is not None:
            max_f0 = self.f0_ceil

        # Workaround segfault issues of WORLD's CheapTrick
        min_f0 = min(min_f0, 500)

        x, fs = sf.read(wav_path)
        assert np.max(x) <= 1.0
        assert x.dtype == np.float64

        if fs != self.sample_rate:
            x = librosa.resample(
                x, orig_sr=fs, target_sr=self.sample_rate, res_type=self.res_type
            )
            if self.sample_rate > fs:
                # NOTE: add a tiny white noise to compensate missing frequency bands
                init_seed(len(x))
                x = x + np.random.randn(len(x)) * 1e-7
            fs = self.sample_rate

        if self.f0_extractor == "parselmouth":
            import parselmouth

            assert (
                self.f0_floor is not None and self.f0_ceil is not None
            ), "must be set manually"
            harvest_num_frames = int(int(1000 * len(x) / fs) / self.frame_period) + 1
            f0 = (
                parselmouth.Sound(x.astype(np.float64), fs)
                .to_pitch_ac(
                    time_step=self.frame_period * 0.001,
                    voicing_threshold=0.6,
                    very_accurate=False,
                    pitch_floor=min_f0,
                    pitch_ceiling=max_f0,
                )
                .selected_array["frequency"]
            )
            pad = int(np.round((3 / min_f0) / (self.frame_period * 0.001)))
            f0 = np.pad(f0, [[0, pad]], mode="constant")
            if len(f0) > harvest_num_frames:
                f0 = f0[:harvest_num_frames]
            elif len(f0) < harvest_num_frames:
                f0 = np.pad(f0, (0, harvest_num_frames - len(f0)), mode="constant")

            timeaxis = np.arange(harvest_num_frames) * self.frame_period * 0.001
        elif self.f0_extractor == "harvest":
            f0, timeaxis = pyworld.harvest(
                x, fs, frame_period=self.frame_period, f0_floor=min_f0, f0_ceil=max_f0
            )
        elif self.f0_extractor == "dio":
            f0, timeaxis = pyworld.dio(
                x, fs, frame_period=self.frame_period, f0_floor=min_f0, f0_ceil=max_f0
            )
            f0 = pyworld.stonemask(x, f0, timeaxis, fs)
        else:
            raise ValueError(f"unknown f0 extractor: {self.f0_extractor}")

        # Workaround for https://github.com/r9y9/nnsvs/issues/7
        f0 = np.maximum(f0, 0)

        # Correct V/UV (and F0) based on the musical score information
        # treat frames where musical notes are not assigned as unvoiced
        # Use smoothed mask so that we don't mask out overshoot or something
        # that could happen at the start/end of notes
        # 0.5 sec. window (could be tuned for better results)
        if self.correct_vuv:
            win_length = int(0.5 / (self.frame_period * 0.001))
            mask = np.convolve(f0_score, np.ones(win_length) / win_length, "same")
            if len(f0) > len(mask):
                mask = np.pad(mask, (0, len(f0) - len(mask)), "constant")
            elif len(f0) < len(mask):
                mask = mask[: len(f0)]
            f0 = f0 * np.sign(mask)

        aperiodicity = pyworld.d4c(x, f0, timeaxis, fs, threshold=self.d4c_threshold)

        if np.isnan(aperiodicity).any():
            print(wav_path)
            print(min_f0, max_f0, aperiodicity.shape, fs)
            print(np.isnan(aperiodicity).sum())
            print(aperiodicity)
            raise RuntimeError("Aperiodicity has NaN")

        # Apply pitch correction
        # NOTE: we should better to apply pitch corrections
        # or manually adjust UST/musicxml in advance.
        sr_f0 = int(1 / (self.frame_period * 0.001))
        if self.correct_f0:
            f0_smooth = extract_smoothed_f0(f0, sr_f0, cutoff=20)
            ratio = compute_f0_correction_ratio(f0_smooth, f0_score)
            f0 *= ratio

        lf0 = f0[:, np.newaxis].copy()
        nonzero_indices = np.nonzero(lf0)
        lf0[nonzero_indices] = np.log(f0[:, np.newaxis][nonzero_indices])
        if self.f0_extractor == "harvest":
            # https://github.com/mmorise/World/issues/35#issuecomment-306521887
            vuv = (aperiodicity[:, 0] < 0.5).astype(np.float32)[:, None]
        else:
            vuv = (lf0 != 0).astype(np.float32)

        # F0 -> continuous F0
        lf0 = interp1d(lf0, kind="slinear")

        # Smooth continuous F0 to avoid discontinuities
        if self.trajectory_smoothing_f0:
            lf0 = extract_smoothed_continuous_f0(
                lf0, sr_f0, cutoff=self.trajectory_smoothing_cutoff_f0
            )

        # Fill continuous F0s for segments where no notes are assigned & no F0s are detected.
        lf0_score = _midi_to_hz(l_features, self.pitch_idx, True)
        clf0_score = interp1d(lf0_score, kind="slinear")
        mask = np.convolve(lf0_score, np.ones(1), "same")
        if len(f0) > len(mask):
            mask = np.pad(mask, (0, len(f0) - len(mask)), "constant")
            clf0_score = np.pad(clf0_score, (0, len(f0) - len(clf0_score)), "constant")
        elif len(f0) < len(mask):
            mask = mask[: len(f0)]
            clf0_score = clf0_score[: len(f0)]
        ind = (mask + f0.reshape(-1)) <= 0
        lf0[ind, 0] = clf0_score[ind]

        # Mel-spectrogram
        logmel = logmelfilterbank(
            x,
            fs,
            fft_size=self.fft_size,
            hop_size=self.hop_size,
            win_length=self.win_length,
            window="hann",
            fmin=self.fmin,
            fmax=self.fmax,
            eps=self.eps,
            num_mels=self.num_mels,
        )

        # Adjust lengths
        logmel = logmel[:num_frames]
        lf0 = lf0[:num_frames]
        vuv = vuv[:num_frames]

        # Concat features
        features = np.hstack((logmel, lf0, vuv)).astype(np.float32)
        pf_features = features

        if len(features) < num_frames:
            print(f"WARN: length mismatch for {wav_path}")
            print(len(features), num_frames)
            return None, None, None

        # Align waveform and features
        wave = x.astype(np.float32)

        # NOTE: since neural vocoders need to perform integer-valued up-sampling
        # (e.g., 120x upsampling with 5ms and 24kHz sampling), we must ensure
        # that the length of the waveform is a multiple of the integer-valued
        # up-sampling factor.
        frame_shift_int = int(fs * self.frame_period / 1000)
        T = int(features.shape[0] * frame_shift_int)
        if len(wave) < T:
            if T - len(wave) > int(fs * (self.frame_period * 0.001)):
                print("Length mismatch", T, len(wave), T - len(wave))
                print(f"Unaligned data: {wav_path} and {label_path}")
                raise RuntimeError("Unaligned data")
            else:
                pass
            wave = np.pad(wave, (0, T - len(wave)))
        assert wave.shape[0] >= T
        wave = wave[:T]

        assert np.isfinite(features).all()
        assert np.isfinite(wave).all()
        assert np.isfinite(pf_features).all()

        return features, wave, pf_features



import numpy as np
import pytest
from nnsvs.dsp import lowpass_filter


@pytest.mark.parametrize("cutoff", [3, 5, 8])
def test_lowpass_filter_noraise(cutoff):
    sr = int(1 / 0.005)

    # special case: 0-length array
    lowpass_filter(np.zeros(10)[0:0], sr, cutoff=cutoff)
    for n in range(sr * 2):
        lowpass_filter(np.random.rand(n), sr, cutoff=cutoff)



# -*- coding: utf-8 -*-

# Copyright 2022 Reo Yoneyama (Nagoya University)
#  MIT License (https://opensource.org/licenses/MIT)

"""Unified Source-Filter GAN Generator modules."""

from logging import getLogger

import torch
import torch.nn as nn
from nnsvs.usfgan.layers import Conv1d1x1, ResidualBlocks, upsample
from nnsvs.usfgan.layers.residual_block import PeriodicityEstimator
from nnsvs.usfgan.utils import index_initial

# A logger for this file
logger = getLogger(__name__)


class USFGANGenerator(nn.Module):
    """Unified Source-Filter GAN Generator module."""

    def __init__(
        self,
        source_network_params={
            "blockA": 30,
            "cycleA": 3,
            "blockF": 0,
            "cycleF": 0,
            "cascade_mode": 0,
        },
        filter_network_params={
            "blockA": 0,
            "cycleA": 0,
            "blockF": 30,
            "cycleF": 3,
            "cascade_mode": 0,
        },
        in_channels=1,
        out_channels=1,
        residual_channels=64,
        gate_channels=128,
        skip_channels=64,
        aux_channels=80,
        aux_context_window=2,
        use_weight_norm=True,
        upsample_params={"upsample_scales": [5, 4, 3, 2]},
    ):
        """Initialize USFGANGenerator module.

        Args:
            source_network_params (dict): Source-network parameters.
            filter_network_params (dict): Filter-network parameters.
            in_channels (int): Number of input channels.
            out_channels (int): Number of output channels.
            residual_channels (int): Number of channels in residual conv.
            gate_channels (int):  Number of channels in gated conv.
            skip_channels (int): Number of channels in skip conv.
            aux_channels (int): Number of channels for auxiliary feature conv.
            aux_context_window (int): Context window size for auxiliary feature.
            use_weight_norm (bool): Whether to use weight norm.
                If set to true, it will be applied to all of the conv layers.
            upsample_params (dict): Upsampling network parameters.

        """
        super().__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.aux_channels = aux_channels
        self.n_ch = residual_channels

        # define first convolution
        self.conv_first = Conv1d1x1(in_channels, residual_channels)

        # define upsampling network
        self.upsample_net = getattr(upsample, "ConvInUpsampleNetwork")(
            **upsample_params,
            aux_channels=aux_channels,
            aux_context_window=aux_context_window,
        )

        # define source/filter networks
        for params in [
            source_network_params,
            filter_network_params,
        ]:
            params.update(
                {
                    "residual_channels": residual_channels,
                    "gate_channels": gate_channels,
                    "skip_channels": skip_channels,
                    "aux_channels": aux_channels,
                }
            )
        self.source_network = ResidualBlocks(**source_network_params)
        self.filter_network = ResidualBlocks(**filter_network_params)

        # convert source signal to hidden representation
        self.conv_mid = Conv1d1x1(out_channels, skip_channels)

        # convert hidden representation to output signal
        self.conv_last = nn.Sequential(
            nn.ReLU(),
            Conv1d1x1(skip_channels, skip_channels),
            nn.ReLU(),
            Conv1d1x1(skip_channels, out_channels),
        )

        # apply weight norm
        if use_weight_norm:
            self.apply_weight_norm()

    def forward(self, x, c, d):
        """Calculate forward propagation.

        Args:
            x (Tensor): Input noise signal (B, 1, T).
            c (Tensor): Local conditioning auxiliary features (B, C ,T').
            d (Tensor): Input pitch-dependent dilated factors (B, 1, T).

        Returns:
            Tensor: Output tensor (B, 1, T)

        """
        # index initialization
        batch_index, ch_index = index_initial(x.size(0), self.n_ch)

        # perform upsampling
        c = self.upsample_net(c)
        assert c.size(-1) == x.size(-1)

        # encode to hidden representation
        x = self.conv_first(x)

        # source excitation generation
        x = self.source_network(x, c, d, batch_index, ch_index)
        s = self.conv_last(x)
        x = self.conv_mid(s)

        # resonance filtering
        x = self.filter_network(x, c, d, batch_index, ch_index)
        x = self.conv_last(x)

        return x, s

    def remove_weight_norm(self):
        """Remove weight normalization module from all of the layers."""

        def _remove_weight_norm(m):
            try:
                logger.debug(f"Weight norm is removed from {m}.")
                nn.utils.remove_weight_norm(m)
            except ValueError:  # this module didn't have weight norm
                return

        self.apply(_remove_weight_norm)

    def apply_weight_norm(self):
        """Apply weight normalization module from all of the layers."""

        def _apply_weight_norm(m):
            if isinstance(m, nn.Conv1d) or isinstance(m, nn.Conv2d):
                nn.utils.weight_norm(m)
                logger.debug(f"Weight norm is applied to {m}.")

        self.apply(_apply_weight_norm)


class CascadeHnUSFGANGenerator(nn.Module):
    """Cascade hn-uSFGAN Generator module."""

    def __init__(
        self,
        harmonic_network_params={
            "blockA": 20,
            "cycleA": 4,
            "blockF": 0,
            "cycleF": 0,
            "cascade_mode": 0,
        },
        noise_network_params={
            "blockA": 0,
            "cycleA": 0,
            "blockF": 5,
            "cycleF": 5,
            "cascade_mode": 0,
        },
        filter_network_params={
            "blockA": 0,
            "cycleA": 0,
            "blockF": 30,
            "cycleF": 3,
            "cascade_mode": 0,
        },
        periodicity_estimator_params={
            "conv_blocks": 3,
            "kernel_size": 5,
            "dilation": 1,
            "padding_mode": "replicate",
        },
        in_channels=1,
        out_channels=1,
        residual_channels=64,
        gate_channels=128,
        skip_channels=64,
        aux_channels=80,
        aux_context_window=2,
        use_weight_norm=True,
        upsample_params={"upsample_scales": [5, 4, 3, 2]},
    ):
        """Initialize CascadeHnUSFGANGenerator module.

        Args:
            harmonic_network_params (dict): Periodic source generation network parameters.
            noise_network_params (dict): Aperiodic source generation network parameters.
            filter_network_params (dict): Filter network parameters.
            periodicity_estimator_params (dict): Periodicity estimation network parameters.
            in_channels (int): Number of input channels.
            out_channels (int): Number of output channels.
            residual_channels (int): Number of channels in residual conv.
            gate_channels (int):  Number of channels in gated conv.
            skip_channels (int): Number of channels in skip conv.
            aux_channels (int): Number of channels for auxiliary feature conv.
            aux_context_window (int): Context window size for auxiliary feature.
            use_weight_norm (bool): Whether to use weight norm.
                If set to true, it will be applied to all of the conv layers.
            upsample_params (dict): Upsampling network parameters.

        """
        super(CascadeHnUSFGANGenerator, self).__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.aux_channels = aux_channels
        self.n_ch = residual_channels

        # define first convolution
        self.conv_first_sine = Conv1d1x1(in_channels, residual_channels)
        self.conv_first_noise = Conv1d1x1(in_channels, residual_channels)
        self.conv_merge = Conv1d1x1(residual_channels * 2, residual_channels)

        # define upsampling network
        self.upsample_net = getattr(upsample, "ConvInUpsampleNetwork")(
            **upsample_params,
            aux_channels=aux_channels,
            aux_context_window=aux_context_window,
        )

        # define harmonic/noise/filter networks
        for params in [
            harmonic_network_params,
            noise_network_params,
            filter_network_params,
        ]:
            params.update(
                {
                    "residual_channels": residual_channels,
                    "gate_channels": gate_channels,
                    "skip_channels": skip_channels,
                    "aux_channels": aux_channels,
                }
            )
        self.harmonic_network = ResidualBlocks(**harmonic_network_params)
        self.noise_network = ResidualBlocks(**noise_network_params)
        self.filter_network = ResidualBlocks(**filter_network_params)

        # define periodicity estimator
        self.periodicity_estimator = PeriodicityEstimator(
            **periodicity_estimator_params, in_channels=aux_channels
        )

        # convert hidden representation to output signal
        self.conv_last = nn.Sequential(
            nn.ReLU(),
            Conv1d1x1(skip_channels, skip_channels),
            nn.ReLU(),
            Conv1d1x1(skip_channels, out_channels),
        )

        # apply weight norm
        if use_weight_norm:
            self.apply_weight_norm()

    def forward(self, x, c, d):
        """Calculate forward propagation.

        Args:
            x (Tensor): Input noise signal (B, 1, T).
            c (Tensor): Local conditioning auxiliary features (B, C ,T').
            d (Tensor): Input pitch-dependent dilated factors (B, 1, T).

        Returns:
            Tensor: Output tensor (B, 1, T)

        """
        # index initialization
        batch_index, ch_index = index_initial(x.size(0), self.n_ch)

        # upsample auxiliary features
        c = self.upsample_net(c)
        assert c.size(-1) == x.size(-1)

        # estimate periodicity
        a = self.periodicity_estimator(c)

        # assume the first channel is sine and the other is noise
        sine, noise = torch.chunk(x, 2, 1)

        # encode to hidden representation
        h = self.conv_first_sine(sine)
        n = self.conv_first_noise(noise)

        # generate periodic and aperiodic source latent features
        h = self.harmonic_network(h, c, d, batch_index, ch_index)
        h = a * h
        n = self.conv_merge(torch.cat([h, n], dim=1))
        n = self.noise_network(n, c, d, batch_index, ch_index)
        n = (1.0 - a) * n

        # merge periodic and aperiodic latent features
        s = h + n

        # resonance filtering
        x = self.filter_network(s, c, d, batch_index, ch_index)
        x = self.conv_last(x)

        # convert to 1d signal for regularization loss
        s = self.conv_last(s)

        # just for debug
        with torch.no_grad():
            h = self.conv_last(h)
            n = self.conv_last(n)

        return x, s, h, n, a

    def remove_weight_norm(self):
        """Remove weight normalization module from all of the layers."""

        def _remove_weight_norm(m):
            try:
                logger.debug(f"Weight norm is removed from {m}.")
                nn.utils.remove_weight_norm(m)
            except ValueError:  # this module didn't have weight norm
                return

        self.apply(_remove_weight_norm)

    def apply_weight_norm(self):
        """Apply weight normalization module from all of the layers."""

        def _apply_weight_norm(m):
            if isinstance(m, nn.Conv1d) or isinstance(m, nn.Conv2d):
                nn.utils.weight_norm(m)
                logger.debug(f"Weight norm is applied to {m}.")

        self.apply(_apply_weight_norm)


class ParallelHnUSFGANGenerator(nn.Module):
    """Parallel hn-uSFGAN Generator module."""

    def __init__(
        self,
        harmonic_network_params={
            "blockA": 20,
            "cycleA": 4,
            "blockF": 0,
            "cycleF": 0,
            "cascade_mode": 0,
        },
        noise_network_params={
            "blockA": 0,
            "cycleA": 0,
            "blockF": 5,
            "cycleF": 5,
            "cascade_mode": 0,
        },
        filter_network_params={
            "blockA": 0,
            "cycleA": 0,
            "blockF": 30,
            "cycleF": 3,
            "cascade_mode": 0,
        },
        periodicity_estimator_params={
            "conv_blocks": 3,
            "kernel_size": 5,
            "dilation": 1,
            "padding_mode": "replicate",
        },
        in_channels=1,
        out_channels=1,
        residual_channels=64,
        gate_channels=128,
        skip_channels=64,
        aux_channels=80,
        aux_context_window=2,
        use_weight_norm=True,
        upsample_params={"upsample_scales": [5, 4, 3, 2]},
    ):
        """Initialize ParallelHnUSFGANGenerator module.

        Args:
            harmonic_network_params (dict): Periodic source generation network parameters.
            noise_network_params (dict): Aperiodic source generation network parameters.
            filter_network_params (dict): Filter network parameters.
            periodicity_estimator_params (dict): Periodicity estimation network parameters.
            in_channels (int): Number of input channels.
            out_channels (int): Number of output channels.
            residual_channels (int): Number of channels in residual conv.
            gate_channels (int):  Number of channels in gated conv.
            skip_channels (int): Number of channels in skip conv.
            aux_channels (int): Number of channels for auxiliary feature conv.
            aux_context_window (int): Context window size for auxiliary feature.
            use_weight_norm (bool): Whether to use weight norm.
                If set to true, it will be applied to all of the conv layers.
            upsample_params (dict): Upsampling network parameters.

        """
        super(ParallelHnUSFGANGenerator, self).__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.aux_channels = aux_channels
        self.n_ch = residual_channels

        # define first convolution
        self.conv_first_sine = Conv1d1x1(in_channels, residual_channels)
        self.conv_first_noise = Conv1d1x1(in_channels, residual_channels)

        # define upsampling network
        self.upsample_net = getattr(upsample, "ConvInUpsampleNetwork")(
            **upsample_params,
            aux_channels=aux_channels,
            aux_context_window=aux_context_window,
        )

        # define harmonic/noise/filter networks
        for params in [
            harmonic_network_params,
            noise_network_params,
            filter_network_params,
        ]:
            params.update(
                {
                    "residual_channels": residual_channels,
                    "gate_channels": gate_channels,
                    "skip_channels": skip_channels,
                    "aux_channels": aux_channels,
                }
            )
        self.harmonic_network = ResidualBlocks(**harmonic_network_params)
        self.noise_network = ResidualBlocks(**noise_network_params)
        self.filter_network = ResidualBlocks(**filter_network_params)

        # define periodicity estimator
        self.periodicity_estimator = PeriodicityEstimator(
            **periodicity_estimator_params, in_channels=aux_channels
        )

        # convert hidden representation to output signal
        self.conv_last = nn.Sequential(
            nn.ReLU(),
            Conv1d1x1(skip_channels, skip_channels),
            nn.ReLU(),
            Conv1d1x1(skip_channels, out_channels),
        )

        # apply weight norm
        if use_weight_norm:
            self.apply_weight_norm()

    def forward(self, x, c, d):
        """Calculate forward propagation.

        Args:
            x (Tensor): Input noise signal (B, 1, T).
            c (Tensor): Local conditioning auxiliary features (B, C ,T').
            d (Tensor): Input pitch-dependent dilated factors (B, 1, T).

        Returns:
            Tensor: Output tensor (B, 1, T)

        """
        # index initialization
        batch_index, ch_index = index_initial(x.size(0), self.n_ch)

        # upsample auxiliary features
        c = self.upsample_net(c)
        assert c.size(-1) == x.size(-1)

        # estimate periodicity
        a = self.periodicity_estimator(c)

        # assume the first channel is sine and the other is noise
        sine, noise = torch.chunk(x, 2, 1)

        # encode to hidden representation
        h = self.conv_first_sine(sine)
        n = self.conv_first_noise(noise)

        # generate periodic and aperiodic source latent features
        h = self.harmonic_network(h, c, d, batch_index, ch_index)
        n = self.noise_network(n, c, d, batch_index, ch_index)

        # merge periodic and aperiodic latent features
        h = a * h
        n = (1.0 - a) * n
        s = h + n

        # resonance filtering
        x = self.filter_network(s, c, d, batch_index, ch_index)
        x = self.conv_last(x)

        # convert to 1d signal for regularization loss
        s = self.conv_last(s)

        # just for debug
        with torch.no_grad():
            h = self.conv_last(h)
            n = self.conv_last(n)

        return x, s, h, n, a

    def remove_weight_norm(self):
        """Remove weight normalization module from all of the layers."""

        def _remove_weight_norm(m):
            try:
                logger.debug(f"Weight norm is removed from {m}.")
                nn.utils.remove_weight_norm(m)
            except ValueError:  # this module didn't have weight norm
                return

        self.apply(_remove_weight_norm)

    def apply_weight_norm(self):
        """Apply weight normalization module from all of the layers."""

        def _apply_weight_norm(m):
            if isinstance(m, nn.Conv1d) or isinstance(m, nn.Conv2d):
                nn.utils.weight_norm(m)
                logger.debug(f"Weight norm is applied to {m}.")

        self.apply(_apply_weight_norm)



import tempfile

import pytest
from nnsvs.logger import getLogger


@pytest.mark.parametrize("verbose", [0, 1, 100])
def test_logger(verbose):
    logger = getLogger(verbose=verbose)

    logger.info(f"verbose={verbose}")


@pytest.mark.parametrize("verbose", [0, 1, 100])
def test_logger_filename(verbose):
    try:
        with tempfile.NamedTemporaryFile() as tmp:
            logger = getLogger(verbose=verbose, filename=tmp.name)
        logger.info(f"verbose={verbose}")
    except PermissionError:
        pass



from .data_source import (
    DurationFeatureSource,
    MelF0AcousticSource,
    MusicalLinguisticSource,
    TimeLagFeatureSource,
    WORLDAcousticSource,
)

__all__ = [
    "DurationFeatureSource",
    "MusicalLinguisticSource",
    "TimeLagFeatureSource",
    "WORLDAcousticSource",
    "MelF0AcousticSource",
]



import pytest
import torch
from nnsvs.discriminators import Conv2dD
from nnsvs.util import init_seed


def _test_model_impl(model, in_dim):
    B = 4
    T = 100
    init_seed(B * T)
    x = torch.rand(B, T, in_dim)
    lengths = torch.Tensor([T] * B).long()

    # warmup forward pass
    with torch.no_grad():
        y = model(x, lengths=lengths)
    # should contain multiple outputs
    assert isinstance(y, list)
    # should contain intermediate outputs to compute feature matching loss
    assert isinstance(y[-1], list)


@pytest.mark.parametrize("padding_mode", ["reflect", "zeros"])
def test_conv2d(padding_mode):
    params = {
        "in_dim": 60,
        "channels": 8,
        "kernel_size": (3, 3),
        "padding": (0, 0),
        "padding_mode": padding_mode,
        "last_sigmoid": False,
        "init_type": "none",
    }
    model = Conv2dD(**params)
    _test_model_impl(model, params["in_dim"])



import unittest

import numpy as np
import torch
import torch.optim as optim
from nnsvs import mdn
from nnsvs.util import init_seed
from torch import nn


class MDN(nn.Module):
    def __init__(self, in_dim, hidden_dim, out_dim, num_layers=1, num_gaussians=30):
        super(MDN, self).__init__()
        self.first_linear = nn.Linear(in_dim, hidden_dim)
        self.hidden_layers = nn.ModuleList(
            [nn.Linear(hidden_dim, hidden_dim) for _ in range(num_layers)]
        )
        self.tanh = nn.Tanh()
        self.mdn = mdn.MDNLayer(hidden_dim, out_dim, num_gaussians=num_gaussians)

    def forward(self, x, lengths=None):
        out = self.tanh(self.first_linear(x))
        for hl in self.hidden_layers:
            out = self.tanh(hl(out))
        return self.mdn(out)


class TestMDN(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        init_seed(42)

        # generate data
        # Inverse model written in PRML Book p. 273
        # https://www.microsoft.com/en-us/research/people/cmbishop/prml-book/
        n = 2500
        self.d_in = 1
        self.d_out = 1

        x_train = np.random.uniform(0, 1, (n, self.d_in)).astype(np.float32)
        noise = np.random.uniform(-0.1, 0.1, (n, self.d_in)).astype(np.float32)
        y_train = x_train + 0.3 * np.sin(2 * np.pi * x_train) + noise

        self.x_train_inv = y_train
        self.y_train_inv = x_train

        self.x_test = np.array([0.0, 0.2, 0.5, 0.8, 1.0]).astype(np.float32)

        # [lower_limit, upper_limit] corresponding to x_test
        self.y_test_range = np.array(
            [[-0.5, 1], [-0.5, 2.0], [0.2, 0.9], [0.8, 1.0], [0.85, 1.05]]
        ).astype(np.float32)

        hidden_dim = 50
        num_gaussians = 30
        num_layers = 0

        self.batch_size = n

        use_cuda = torch.cuda.is_available()

        self.device = torch.device("cuda" if use_cuda else "cpu")
        self.model = MDN(
            self.d_in,
            hidden_dim,
            self.d_out,
            num_layers=num_layers,
            num_gaussians=num_gaussians,
        ).to(self.device)

        learning_rate = 0.008
        self.opt = optim.Adam(self.model.parameters(), lr=learning_rate)

    def test_mdn_loss(self):
        # wrap up the inverse data as Variables
        x = torch.from_numpy(
            self.x_train_inv.reshape(self.batch_size, -1, self.d_in)
        ).to(
            self.device
        )  # (B, max(T), D_in)
        y = torch.from_numpy(
            self.y_train_inv.reshape(self.batch_size, -1, self.d_out)
        ).to(
            self.device
        )  # (B, max(T), D_out)
        for e in range(1000):
            self.model.zero_grad()
            pi, sigma, mu = self.model(x)
            loss = mdn.mdn_loss(pi, sigma, mu, y).mean()
            if e % 100 == 0:
                print(f"loss: {loss.data.item()}")
            loss.backward()
            self.opt.step()

    def test_mdn_get_most_probable_sigma_and_mu(self):
        self.test_mdn_loss()

        pi, sigma, mu = self.model(
            torch.from_numpy(self.x_test.reshape(1, -1, self.d_in)).to(self.device)
        )
        _, max_mu = mdn.mdn_get_most_probable_sigma_and_mu(pi, sigma, mu)
        max_mu = max_mu.squeeze(0).cpu().detach().numpy()
        print(max_mu.shape)

        for i, sample in enumerate(max_mu):
            lower_limit = self.y_test_range[i][0]
            upper_limit = self.y_test_range[i][1]
            assert lower_limit < sample and upper_limit > sample
            print(
                f"sample: {sample}, lower_limit: {lower_limit}, upper_limit: {upper_limit}"
            )

    def test_mdn_get_sample(self):
        self.test_mdn_loss()

        pi, sigma, mu = self.model(
            torch.from_numpy(self.x_test.reshape(1, -1, self.d_in)).to(self.device)
        )
        samples = mdn.mdn_get_sample(pi, sigma, mu).squeeze(0).cpu().detach().numpy()

        for i, sample in enumerate(samples):
            lower_limit = self.y_test_range[i][0]
            upper_limit = self.y_test_range[i][1]
            assert lower_limit < sample and upper_limit > sample
            print(
                f"sample: {sample}, lower_limit: {lower_limit}, upper_limit: {upper_limit}"
            )


if __name__ == "__main__":
    unittest.main()



import math

import torch
from torch import nn


def denorm_f0(
    f0,
    uv,
    pitch_padding=None,
    min=None,
    max=None,
    pitch_norm="log",
    use_uv=True,
    f0_std=1.0,
    f0_mean=0.0,
):
    assert use_uv

    if pitch_norm == "standard":
        f0 = f0 * f0_std + f0_mean
    elif pitch_norm == "log":
        f0 = 2 ** f0

    if min is not None:
        f0 = f0.clamp(min=min)
    if max is not None:
        f0 = f0.clamp(max=max)
    if uv is not None and use_uv:
        f0[uv > 0] = 0
    if pitch_padding is not None:
        f0[pitch_padding] = 0
    return f0


def LayerNorm(normalized_shape, eps=1e-5, elementwise_affine=True, export=False):
    if not export and torch.cuda.is_available():
        try:
            from apex.normalization import FusedLayerNorm

            return FusedLayerNorm(normalized_shape, eps, elementwise_affine)
        except ImportError:
            pass
    return torch.nn.LayerNorm(normalized_shape, eps, elementwise_affine)


class ConvNorm(torch.nn.Module):
    def __init__(
        self,
        in_channels,
        out_channels,
        kernel_size=1,
        stride=1,
        padding=None,
        dilation=1,
        bias=True,
        w_init_gain="linear",
    ):
        super(ConvNorm, self).__init__()
        if padding is None:
            assert kernel_size % 2 == 1
            padding = int(dilation * (kernel_size - 1) / 2)

        self.conv = torch.nn.Conv1d(
            in_channels,
            out_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            dilation=dilation,
            bias=bias,
        )

        torch.nn.init.xavier_uniform_(
            self.conv.weight, gain=torch.nn.init.calculate_gain(w_init_gain)
        )

    def forward(self, signal):
        conv_signal = self.conv(signal)
        return conv_signal


def Linear(in_features, out_features, bias=True):
    m = nn.Linear(in_features, out_features, bias)
    nn.init.xavier_uniform_(m.weight)
    if bias:
        nn.init.constant_(m.bias, 0.0)
    return m


def make_positions(tensor, padding_idx):
    """Replace non-padding symbols with their position numbers.

    Position numbers begin at padding_idx+1. Padding symbols are ignored.
    """
    # The series of casts and type-conversions here are carefully
    # balanced to both work with ONNX export and XLA. In particular XLA
    # prefers ints, cumsum defaults to output longs, and ONNX doesn't know
    # how to handle the dtype kwarg in cumsum.
    mask = tensor.ne(padding_idx).int()
    return (torch.cumsum(mask, dim=1).type_as(mask) * mask).long() + padding_idx


class SinusoidalPositionalEmbedding(nn.Module):
    """This module produces sinusoidal positional embeddings of any length.

    Padding symbols are ignored.
    """

    def __init__(self, embedding_dim, padding_idx, init_size=1024):
        super().__init__()
        self.embedding_dim = embedding_dim
        self.padding_idx = padding_idx
        self.weights = SinusoidalPositionalEmbedding.get_embedding(
            init_size,
            embedding_dim,
            padding_idx,
        )
        self.register_buffer("_float_tensor", torch.FloatTensor(1))

    @staticmethod
    def get_embedding(num_embeddings, embedding_dim, padding_idx=None):
        """Build sinusoidal embeddings.

        This matches the implementation in tensor2tensor, but differs slightly
        from the description in Section 3.5 of "Attention Is All You Need".
        """
        half_dim = embedding_dim // 2
        emb = math.log(10000) / (half_dim - 1)
        emb = torch.exp(torch.arange(half_dim, dtype=torch.float) * -emb)
        emb = torch.arange(num_embeddings, dtype=torch.float).unsqueeze(
            1
        ) * emb.unsqueeze(0)
        emb = torch.cat([torch.sin(emb), torch.cos(emb)], dim=1).view(
            num_embeddings, -1
        )
        if embedding_dim % 2 == 1:
            # zero pad
            emb = torch.cat([emb, torch.zeros(num_embeddings, 1)], dim=1)
        if padding_idx is not None:
            emb[padding_idx, :] = 0
        return emb

    def forward(
        self, input, incremental_state=None, timestep=None, positions=None, **kwargs
    ):
        """Input is expected to be of size [bsz x seqlen]."""
        bsz, seq_len = input.shape[:2]
        max_pos = self.padding_idx + 1 + seq_len
        if self.weights is None or max_pos > self.weights.size(0):
            # recompute/expand embeddings if needed
            self.weights = SinusoidalPositionalEmbedding.get_embedding(
                max_pos,
                self.embedding_dim,
                self.padding_idx,
            )
        self.weights = self.weights.to(self._float_tensor)

        if incremental_state is not None:
            # positions is the same for every token when decoding a single step
            pos = timestep.view(-1)[0] + 1 if timestep is not None else seq_len
            return self.weights[self.padding_idx + pos, :].expand(bsz, 1, -1)

        positions = (
            make_positions(input, self.padding_idx) if positions is None else positions
        )
        return (
            self.weights.index_select(0, positions.view(-1))
            .view(bsz, seq_len, -1)
            .detach()
        )

    def max_positions(self):
        """Maximum number of supported positions."""
        return int(1e5)  # an arbitrary large number


class FSLayerNorm(torch.nn.LayerNorm):
    """Layer normalization module.
    :param int nout: output dim size
    :param int dim: dimension to be normalized
    """

    def __init__(self, nout, dim=-1):
        """Construct an LayerNorm object."""
        super(FSLayerNorm, self).__init__(nout, eps=1e-12)
        self.dim = dim

    def forward(self, x):
        """Apply layer normalization.
        :param torch.Tensor x: input tensor
        :return: layer normalized tensor
        :rtype torch.Tensor
        """
        if self.dim == -1:
            return super(FSLayerNorm, self).forward(x)
        return super(FSLayerNorm, self).forward(x.transpose(1, -1)).transpose(1, -1)


class PitchPredictor(torch.nn.Module):
    def __init__(
        self,
        idim,
        n_layers=5,
        n_chans=384,
        odim=2,
        kernel_size=5,
        dropout_rate=0.1,
        padding="SAME",
    ):
        """Initialize pitch predictor module.
        Args:
            idim (int): Input dimension.
            n_layers (int, optional): Number of convolutional layers.
            n_chans (int, optional): Number of channels of convolutional layers.
            kernel_size (int, optional): Kernel size of convolutional layers.
            dropout_rate (float, optional): Dropout rate.
        """
        super(PitchPredictor, self).__init__()
        self.conv = torch.nn.ModuleList()
        self.kernel_size = kernel_size
        self.padding = padding
        for idx in range(n_layers):
            in_chans = idim if idx == 0 else n_chans
            self.conv += [
                torch.nn.Sequential(
                    torch.nn.ConstantPad1d(
                        ((kernel_size - 1) // 2, (kernel_size - 1) // 2)
                        if padding == "SAME"
                        else (kernel_size - 1, 0),
                        0,
                    ),
                    torch.nn.Conv1d(
                        in_chans, n_chans, kernel_size, stride=1, padding=0
                    ),
                    torch.nn.ReLU(),
                    FSLayerNorm(n_chans, dim=1),
                    torch.nn.Dropout(dropout_rate),
                )
            ]
        self.linear = torch.nn.Linear(n_chans, odim)
        self.embed_positions = SinusoidalPositionalEmbedding(idim, 0, init_size=4096)
        self.pos_embed_alpha = nn.Parameter(torch.Tensor([1]))

    def forward(self, xs):
        """

        :param xs: [B, T, H]
        :return: [B, T, H]
        """
        positions = self.pos_embed_alpha * self.embed_positions(xs[..., 0])
        xs = xs + positions
        xs = xs.transpose(1, -1)  # (B, idim, Tmax)
        for f in self.conv:
            xs = f(xs)  # (B, C, Tmax)
        # NOTE: calculate in log domain
        xs = self.linear(xs.transpose(1, -1))  # (B, Tmax, H)
        return xs


class Prenet(nn.Module):
    def __init__(self, in_dim=80, out_dim=256, kernel=5, n_layers=3, strides=None):
        super(Prenet, self).__init__()
        padding = kernel // 2
        self.layers = []
        self.strides = strides if strides is not None else [1] * n_layers
        for idx in range(n_layers):
            self.layers.append(
                nn.Sequential(
                    nn.Conv1d(
                        in_dim,
                        out_dim,
                        kernel_size=kernel,
                        padding=padding,
                        stride=self.strides[idx],
                    ),
                    nn.ReLU(),
                    nn.BatchNorm1d(out_dim),
                )
            )
            in_dim = out_dim
        self.layers = nn.ModuleList(self.layers)
        self.out_proj = nn.Linear(out_dim, out_dim)

    def forward(self, x):
        """

        :param x: [B, T, 80]
        :return: [L, B, T, H], [B, T, H]
        """
        padding_mask = x.abs().sum(-1).eq(0).data  # [B, T]
        nonpadding_mask_TB = 1 - padding_mask.float()[:, None, :]  # [B, 1, T]
        x = x.transpose(1, 2)
        hiddens = []
        for i, l in enumerate(self.layers):
            nonpadding_mask_TB = nonpadding_mask_TB[:, :, :: self.strides[i]]
            x = l(x) * nonpadding_mask_TB
        hiddens.append(x)
        hiddens = torch.stack(hiddens, 0)  # [L, B, H, T]
        hiddens = hiddens.transpose(2, 3)  # [L, B, T, H]
        x = self.out_proj(x.transpose(1, 2))  # [B, T, H]
        x = x * nonpadding_mask_TB.transpose(1, 2)
        return hiddens, x


class ConvBlock(nn.Module):
    def __init__(
        self, idim=80, n_chans=256, kernel_size=3, stride=1, norm="gn", dropout=0
    ):
        super().__init__()
        self.conv = ConvNorm(idim, n_chans, kernel_size, stride=stride)
        self.norm = norm
        if self.norm == "bn":
            self.norm = nn.BatchNorm1d(n_chans)
        elif self.norm == "in":
            self.norm = nn.InstanceNorm1d(n_chans, affine=True)
        elif self.norm == "gn":
            self.norm = nn.GroupNorm(n_chans // 16, n_chans)
        elif self.norm == "ln":
            self.norm = LayerNorm(n_chans // 16, n_chans)
        elif self.norm == "wn":
            self.conv = torch.nn.utils.weight_norm(self.conv.conv)
        self.dropout = nn.Dropout(dropout)
        self.relu = nn.ReLU()

    def forward(self, x):
        """

        :param x: [B, C, T]
        :return: [B, C, T]
        """
        x = self.conv(x)
        if not isinstance(self.norm, str):
            if self.norm == "none":
                pass
            elif self.norm == "ln":
                x = self.norm(x.transpose(1, 2)).transpose(1, 2)
            else:
                x = self.norm(x)
        x = self.relu(x)
        x = self.dropout(x)
        return x


class ConvStacks(nn.Module):
    def __init__(
        self,
        idim=80,
        n_layers=5,
        n_chans=256,
        odim=32,
        kernel_size=5,
        norm="gn",
        dropout=0,
        strides=None,
        res=True,
    ):
        super().__init__()
        self.conv = torch.nn.ModuleList()
        self.kernel_size = kernel_size
        self.res = res
        self.in_proj = Linear(idim, n_chans)
        if strides is None:
            strides = [1] * n_layers
        else:
            assert len(strides) == n_layers
        for idx in range(n_layers):
            self.conv.append(
                ConvBlock(
                    n_chans,
                    n_chans,
                    kernel_size,
                    stride=strides[idx],
                    norm=norm,
                    dropout=dropout,
                )
            )
        self.out_proj = Linear(n_chans, odim)

    def forward(self, x, return_hiddens=False):
        """

        :param x: [B, T, H]
        :return: [B, T, H]
        """
        x = self.in_proj(x)
        x = x.transpose(1, -1)  # (B, idim, Tmax)
        hiddens = []
        for f in self.conv:
            x_ = f(x)
            x = x + x_ if self.res else x_  # (B, C, Tmax)
            hiddens.append(x)
        x = x.transpose(1, -1)
        x = self.out_proj(x)  # (B, Tmax, H)
        if return_hiddens:
            hiddens = torch.stack(hiddens, 1)  # [B, L, C, T]
            return x, hiddens
        return x


class PitchExtractor(nn.Module):
    def __init__(
        self,
        n_mel_bins=80,
        conv_layers=2,
        hidden_size=256,
        predictor_hidden=-1,
        ffn_padding="SAME",
        predictor_kernel=5,
        pitch_type="frame",
        use_uv=True,
    ):
        super().__init__()
        self.hidden_size = hidden_size
        self.pitch_type = pitch_type
        assert pitch_type == "log"
        self.use_uv = use_uv
        self.predictor_hidden = (
            predictor_hidden if predictor_hidden > 0 else self.hidden_size
        )
        self.conv_layers = conv_layers

        self.mel_prenet = Prenet(n_mel_bins, self.hidden_size, strides=[1, 1, 1])
        if self.conv_layers > 0:
            self.mel_encoder = ConvStacks(
                idim=self.hidden_size,
                n_chans=self.hidden_size,
                odim=self.hidden_size,
                n_layers=self.conv_layers,
            )
        self.pitch_predictor = PitchPredictor(
            self.hidden_size,
            n_chans=self.predictor_hidden,
            n_layers=5,
            dropout_rate=0.1,
            odim=2,
            padding=ffn_padding,
            kernel_size=predictor_kernel,
        )

    def forward(self, mel_input=None):
        mel_hidden = self.mel_prenet(mel_input)[1]
        if self.conv_layers > 0:
            mel_hidden = self.mel_encoder(mel_hidden)

        # log2(f0), uv
        pitch_pred = self.pitch_predictor(mel_hidden)
        lf0, uv = pitch_pred[:, :, 0], pitch_pred[:, :, 1]

        # f0
        f0 = 2 ** lf0

        # log(f0)
        lf0 = torch.log(f0)

        lf0[uv > 0] = 0

        return lf0


class PitchExtractorWrapper(nn.Module):
    def __init__(self, **kwargs):
        super().__init__()
        self.model = PitchExtractor(**kwargs)

    def forward(self, x, lengths=None, y=None):
        return self.pitch_extractor(x)



from os.path import dirname, join

import numpy as np
import pytest
from nnmnkwii.io import hts
from nnsvs.io.hts import (
    full_to_mono,
    get_note_indices,
    label2phrases,
    label2phrases_str,
    merge_sil,
    overwrite_phoneme_flags_,
    segment_labels,
)


def test_full_to_mono():
    full_labels = hts.load(
        join(dirname(__file__), "data", "neutrino", "sample1_full.lab")
    )
    mono_labels = hts.load(
        join(dirname(__file__), "data", "neutrino", "sample1_mono.lab")
    )

    assert str(mono_labels) == str(full_to_mono(full_labels))


@pytest.mark.parametrize("filename", ["sample1", "sample2", "sample3", "meltdown_main"])
def test_get_note_indices(filename):
    mono_labels = hts.load(
        join(dirname(__file__), "data", "neutrino", f"{filename}_mono.lab")
    )
    note_indices1 = get_note_indices(mono_labels)

    full_labels = hts.load(
        join(dirname(__file__), "data", "neutrino", f"{filename}_full.lab")
    )
    note_indices2 = get_note_indices(full_labels)

    assert (np.array(note_indices1) == np.array(note_indices2)).all()


@pytest.mark.parametrize("filename", ["sample1", "sample2", "sample3", "meltdown_main"])
def test_label2phrases_str(filename):
    mono_labels = hts.load(
        join(dirname(__file__), "data", "neutrino", f"{filename}_mono.lab")
    )
    note_indices = get_note_indices(mono_labels)

    timing_labels = hts.load(
        join(dirname(__file__), "data", "neutrino", f"{filename}_timing.lab")
    )
    nnsvs_str = label2phrases_str(timing_labels, note_indices)

    with open(
        join(dirname(__file__), "data", "neutrino", f"{filename}-phraselist.txt"), "r"
    ) as f:
        neutrino_str = f.read()

    assert nnsvs_str == neutrino_str


@pytest.mark.parametrize("filename", ["sample1", "sample2", "sample3", "meltdown_main"])
@pytest.mark.parametrize("fix_offset", [True, False])
def test_label2phrases(filename, fix_offset):
    timing_labels = hts.load(
        join(dirname(__file__), "data", "neutrino", f"{filename}_timing.lab")
    )
    phrases = label2phrases(timing_labels, fix_offset=fix_offset)
    assert len(phrases) > 0
    assert isinstance(phrases, list)
    for idx, phrase in enumerate(phrases):
        print(f"Phrase {idx}:", " ".join(phrase.contexts))
        assert isinstance(phrase, hts.HTSLabelFile)

    # Check if the number of labels are the same
    assert len(timing_labels) == sum(len(phrase) for phrase in phrases)


@pytest.mark.parametrize(
    "filename", ["sample1", "sample2", "sample3", "meltdown_main", "sample4"]
)
def test_segment_label(filename):
    timing_labels = hts.load(
        join(dirname(__file__), "data", "neutrino", f"{filename}_timing.lab")
    )
    timing_labels = merge_sil(timing_labels)
    phrases = segment_labels(timing_labels)
    assert len(phrases) > 0
    assert isinstance(phrases, list)
    for idx, phrase in enumerate(phrases):
        print(f"Phrase {idx}:", " ".join(phrase.contexts))
        assert isinstance(phrase, hts.HTSLabelFile)

    # Check if the number of labels are the same
    assert len(timing_labels) == sum(len(phrase) for phrase in phrases)


def test_overwrite_phoneme_flags():
    labels = hts.HTSLabelFile()
    labels.append((0, 50000, "v@xx^pau-o+r=e_xx%xx^xx_xx~xx-1!1[xx$xx]xx/A:"))
    overwrite_phoneme_flags_(labels, "soft")
    assert labels.contexts[0] == "v@xx^pau-o+r=e_xx%xx^soft_xx~xx-1!1[xx$xx]xx/A:"
    overwrite_phoneme_flags_(labels, "xx")
    assert labels.contexts[0] == "v@xx^pau-o+r=e_xx%xx^xx_xx~xx-1!1[xx$xx]xx/A:"
    overwrite_phoneme_flags_(labels, "loud")
    assert labels.contexts[0] == "v@xx^pau-o+r=e_xx%xx^loud_xx~xx-1!1[xx$xx]xx/A:"



import math
from math import sqrt

import torch
import torch.nn as nn
import torch.nn.functional as F


class Mish(nn.Module):
    def forward(self, x):
        return x * torch.tanh(F.softplus(x))


class SinusoidalPosEmb(nn.Module):
    def __init__(self, dim):
        super().__init__()
        self.dim = dim

    def forward(self, x):
        device = x.device
        half_dim = self.dim // 2
        emb = math.log(10000) / (half_dim - 1)
        emb = torch.exp(torch.arange(half_dim, device=device) * -emb)
        emb = x[:, None] * emb[None, :]
        emb = torch.cat((emb.sin(), emb.cos()), dim=-1)
        return emb


def Conv1d(*args, **kwargs):
    layer = nn.Conv1d(*args, **kwargs)
    nn.init.kaiming_normal_(layer.weight)
    return layer


@torch.jit.script
def silu(x):
    return x * torch.sigmoid(x)


class ResidualBlock(nn.Module):
    def __init__(self, encoder_hidden, residual_channels, dilation):
        super().__init__()
        self.dilated_conv = Conv1d(
            residual_channels,
            2 * residual_channels,
            3,
            padding=dilation,
            dilation=dilation,
        )
        self.diffusion_projection = nn.Linear(residual_channels, residual_channels)
        self.conditioner_projection = Conv1d(encoder_hidden, 2 * residual_channels, 1)
        self.output_projection = Conv1d(residual_channels, 2 * residual_channels, 1)

    def forward(self, x, conditioner, diffusion_step):
        diffusion_step = self.diffusion_projection(diffusion_step).unsqueeze(-1)
        conditioner = self.conditioner_projection(conditioner)
        y = x + diffusion_step

        y = self.dilated_conv(y) + conditioner

        gate, filter = torch.chunk(y, 2, dim=1)
        y = torch.sigmoid(gate) * torch.tanh(filter)

        y = self.output_projection(y)
        residual, skip = torch.chunk(y, 2, dim=1)
        return (x + residual) / sqrt(2.0), skip


class DiffNet(nn.Module):
    def __init__(
        self,
        in_dim=80,
        encoder_hidden_dim=256,
        residual_layers=20,
        residual_channels=256,
        dilation_cycle_length=4,
    ):
        super().__init__()
        self.in_dim = in_dim

        self.input_projection = Conv1d(in_dim, residual_channels, 1)
        self.diffusion_embedding = SinusoidalPosEmb(residual_channels)
        dim = residual_channels
        self.mlp = nn.Sequential(
            nn.Linear(dim, dim * 4), Mish(), nn.Linear(dim * 4, dim)
        )
        self.residual_layers = nn.ModuleList(
            [
                ResidualBlock(
                    encoder_hidden_dim,
                    residual_channels,
                    2 ** (i % dilation_cycle_length),
                )
                for i in range(residual_layers)
            ]
        )
        self.skip_projection = Conv1d(residual_channels, residual_channels, 1)
        self.output_projection = Conv1d(residual_channels, in_dim, 1)
        nn.init.zeros_(self.output_projection.weight)

    def forward(self, spec, diffusion_step, cond):
        """

        :param spec: [B, 1, M, T]
        :param diffusion_step: [B, 1]
        :param cond: [B, M, T]
        :return:
        """
        x = spec[:, 0]
        x = self.input_projection(x)  # x [B, residual_channel, T]

        x = F.relu(x)
        diffusion_step = self.diffusion_embedding(diffusion_step)
        diffusion_step = self.mlp(diffusion_step)
        skip = []
        for _, layer in enumerate(self.residual_layers):
            x, skip_connection = layer(x, cond, diffusion_step)
            skip.append(skip_connection)

        x = torch.sum(torch.stack(skip), dim=0) / sqrt(len(self.residual_layers))
        x = self.skip_projection(x)
        x = F.relu(x)
        x = self.output_projection(x)  # [B, 80, T]
        return x[:, None, :, :]



import torch
from nnsvs.acoustic_models.util import predict_lf0_with_residual
from nnsvs.base import BaseModel, PredictionType
from nnsvs.mdn import MDNLayer, mdn_get_most_probable_sigma_and_mu
from nnsvs.util import init_weights
from torch import nn
from torch.nn.utils.rnn import pack_padded_sequence, pad_packed_sequence

__all__ = [
    "ResSkipF0FFConvLSTM",
]


class ResSkipF0FFConvLSTM(BaseModel):
    """FFN + Conv1d + LSTM + residual/skip connections

    A model proposed in :cite:t:`hono2021sinsy`.

    Args:
        in_dim (int): input dimension
        ff_hidden_dim (int): hidden dimension of feed-forward layer
        conv_hidden_dim (int): hidden dimension of convolutional layer
        lstm_hidden_dim (int): hidden dimension of LSTM layer
        out_dim (int): output dimension
        dropout (float): dropout rate
        num_ls (int): number of layers of LSTM
        bidirectional (bool): whether to use bidirectional LSTM or not
        in_lf0_idx (int): index of lf0 in input features
        in_lf0_min (float): minimum of lf0 in the training data of input features
        in_lf0_max (float): maximum of lf0 in the training data of input features
        out_lf0_idx (int): index of lf0 in output features
        out_lf0_mean (float): mean of lf0 in the training data of output features
        out_lf0_scale (float): scale of lf0 in the training data of output features
        skip_inputs (bool): whether to use skip connection for the input features
        init_type (str): initialization type
        use_mdn (bool): whether to use MDN or not
        num_gaussians (int): number of gaussians in MDN
        dim_wise (bool): whether to use MDN with dim-wise or not
    """

    def __init__(
        self,
        in_dim,
        ff_hidden_dim=2048,
        conv_hidden_dim=1024,
        lstm_hidden_dim=256,
        out_dim=199,
        dropout=0.0,
        num_lstm_layers=2,
        bidirectional=True,
        # NOTE: you must carefully set the following parameters
        in_lf0_idx=300,
        in_lf0_min=5.3936276,
        in_lf0_max=6.491111,
        out_lf0_idx=180,
        out_lf0_mean=5.953093881972361,
        out_lf0_scale=0.23435173188961034,
        skip_inputs=False,
        init_type="none",
        use_mdn=False,
        num_gaussians=8,
        dim_wise=False,
    ):
        super().__init__()
        self.in_lf0_idx = in_lf0_idx
        self.in_lf0_min = in_lf0_min
        self.in_lf0_max = in_lf0_max
        self.out_lf0_idx = out_lf0_idx
        self.out_lf0_mean = out_lf0_mean
        self.out_lf0_scale = out_lf0_scale
        self.skip_inputs = skip_inputs
        self.use_mdn = use_mdn

        self.ff = nn.Sequential(
            nn.Linear(in_dim, ff_hidden_dim),
            nn.ReLU(),
            nn.Linear(ff_hidden_dim, ff_hidden_dim),
            nn.ReLU(),
            nn.Linear(ff_hidden_dim, ff_hidden_dim),
            nn.ReLU(),
        )

        self.conv = nn.Sequential(
            nn.ReflectionPad1d(3),
            nn.Conv1d(ff_hidden_dim + 1, conv_hidden_dim, kernel_size=7, padding=0),
            nn.BatchNorm1d(conv_hidden_dim),
            nn.ReLU(),
            nn.ReflectionPad1d(3),
            nn.Conv1d(conv_hidden_dim, conv_hidden_dim, kernel_size=7, padding=0),
            nn.BatchNorm1d(conv_hidden_dim),
            nn.ReLU(),
            nn.ReflectionPad1d(3),
            nn.Conv1d(conv_hidden_dim, conv_hidden_dim, kernel_size=7, padding=0),
            nn.BatchNorm1d(conv_hidden_dim),
            nn.ReLU(),
        )

        num_direction = 2 if bidirectional else 1
        self.lstm = nn.LSTM(
            conv_hidden_dim,
            lstm_hidden_dim,
            num_lstm_layers,
            bidirectional=True,
            batch_first=True,
            dropout=dropout,
        )

        if self.skip_inputs:
            last_in_dim = num_direction * lstm_hidden_dim + in_dim
        else:
            last_in_dim = num_direction * lstm_hidden_dim

        if self.use_mdn:
            self.mdn_layer = MDNLayer(
                last_in_dim, out_dim, num_gaussians=num_gaussians, dim_wise=dim_wise
            )
        else:
            self.fc = nn.Linear(last_in_dim, out_dim)

        init_weights(self, init_type)

    def prediction_type(self):
        return (
            PredictionType.PROBABILISTIC
            if self.use_mdn
            else PredictionType.DETERMINISTIC
        )

    def has_residual_lf0_prediction(self):
        return True

    def forward(self, x, lengths=None, y=None):
        """Forward step

        Args:
            x (torch.Tensor): the input tensor
            lengths (torch.Tensor): the lengths of the input tensor
            y (torch.Tensor): the optional target tensor

        Returns:
            torch.Tensor: the output tensor
        """
        if isinstance(lengths, torch.Tensor):
            lengths = lengths.to("cpu")

        lf0_score = x[:, :, self.in_lf0_idx].unsqueeze(-1)

        out = self.ff(x)
        out = torch.cat([out, lf0_score], dim=-1)

        out = self.conv(out.transpose(1, 2)).transpose(1, 2)
        sequence = pack_padded_sequence(out, lengths, batch_first=True)
        out, _ = self.lstm(sequence)
        out, _ = pad_packed_sequence(out, batch_first=True)
        out = torch.cat([out, x], dim=-1) if self.skip_inputs else out

        if self.use_mdn:
            log_pi, log_sigma, mu = self.mdn_layer(out)
        else:
            mu = self.fc(out)

        lf0_pred, lf0_residual = predict_lf0_with_residual(
            x,
            mu,
            self.in_lf0_idx,
            self.in_lf0_min,
            self.in_lf0_max,
            self.out_lf0_idx,
            self.out_lf0_mean,
            self.out_lf0_scale,
        )

        # Inject the predicted lf0 into the output features
        if self.use_mdn:
            mu[:, :, :, self.out_lf0_idx] = lf0_pred
        else:
            mu[:, :, self.out_lf0_idx] = lf0_pred.squeeze(-1)

        if self.use_mdn:
            return (log_pi, log_sigma, mu), lf0_residual
        else:
            return mu, lf0_residual

    def inference(self, x, lengths=None):
        """Inference step

        Args:
            x (torch.Tensor): input features
            lengths (torch.Tensor): lengths of input features

        Returns:
            tuple: (mu, sigma) if use_mdn, (output, ) otherwise
        """
        if self.use_mdn:
            (log_pi, log_sigma, mu), _ = self(x, lengths)
            sigma, mu = mdn_get_most_probable_sigma_and_mu(log_pi, log_sigma, mu)
            return mu, sigma
        else:
            return self(x, lengths)[0]



import torch
from nnsvs.base import PredictionType
from nnsvs.model import (
    FFN,
    LSTMRNN,
    LSTMRNNSAR,
    MDN,
    RMDN,
    Conv1dResnet,
    Conv1dResnetMDN,
    Conv1dResnetSAR,
    FFConvLSTM,
    MDNv2,
    VariancePredictor,
)
from nnsvs.util import init_seed


def test_deprecated_imports():
    from nnsvs.model import ResF0Conv1dResnet  # noqa: F401
    from nnsvs.model import ResF0Conv1dResnetMDN  # noqa: F401
    from nnsvs.model import ResF0VariancePredictor  # noqa: F401
    from nnsvs.model import ResSkipF0FFConvLSTM  # noqa: F401


def _test_model_impl(model, in_dim, out_dim):
    B = 4
    T = 100
    init_seed(B * T)
    x = torch.rand(B, T, in_dim)
    lengths = torch.Tensor([T] * B).long()

    # warmup forward pass
    with torch.no_grad():
        y = model(x, lengths)
        y_inf = model.inference(x, lengths)

    # MDN case
    if model.prediction_type() == PredictionType.PROBABILISTIC:
        log_pi, log_sigma, mu = y
        num_gaussian = log_pi.shape[2]
        assert mu.shape == (B, T, num_gaussian, out_dim)
        assert log_sigma.shape == (B, T, num_gaussian, out_dim)

        # NOTE: infernece output shouldn't have num_gaussian axis
        mu_inf, sigma_inf = y_inf
        assert mu_inf.shape == (B, T, out_dim)
        assert sigma_inf.shape == (B, T, out_dim)
    else:
        assert y.shape == (B, T, out_dim)
        assert y.shape == y_inf.shape


def test_ffn():
    params = {
        "in_dim": 300,
        "hidden_dim": 8,
        "out_dim": 20,
        "num_layers": 2,
        "dropout": 0.1,
        "init_type": "none",
    }
    model = FFN(**params)
    assert model.prediction_type() == PredictionType.DETERMINISTIC
    _test_model_impl(model, params["in_dim"], params["out_dim"])


def test_lstmrnn():
    params = {
        "in_dim": 300,
        "hidden_dim": 8,
        "out_dim": 20,
        "num_layers": 2,
        "dropout": 0.1,
        "init_type": "none",
    }
    model = LSTMRNN(**params)
    assert model.prediction_type() == PredictionType.DETERMINISTIC
    _test_model_impl(model, params["in_dim"], params["out_dim"])


def test_conv1d_resnet():
    params = {
        "in_dim": 300,
        "hidden_dim": 8,
        "out_dim": 180,
        "num_layers": 2,
        "num_gaussians": 2,
        "dim_wise": True,
        "init_type": "none",
    }
    model = Conv1dResnet(**{**params, "use_mdn": False})
    assert model.prediction_type() == PredictionType.DETERMINISTIC
    _test_model_impl(model, params["in_dim"], params["out_dim"])

    model = Conv1dResnet(**{**params, "use_mdn": True})
    assert model.prediction_type() == PredictionType.PROBABILISTIC
    _test_model_impl(model, params["in_dim"], params["out_dim"])

    # Deprecated
    model = Conv1dResnetMDN(**params)
    assert model.prediction_type() == PredictionType.PROBABILISTIC
    _test_model_impl(model, params["in_dim"], params["out_dim"])


def test_conv1d_resnet_sar():
    params = {
        "in_dim": 300,
        "hidden_dim": 8,
        "out_dim": 20,
        "num_layers": 2,
        "stream_sizes": [10, 10],
        "ar_orders": [2, 2],
        "init_type": "none",
    }
    model = Conv1dResnetSAR(**params)
    assert model.prediction_type() == PredictionType.DETERMINISTIC
    _test_model_impl(model, params["in_dim"], params["out_dim"])

    y = torch.rand(8, 100, params["out_dim"])
    y_hat = model.preprocess_target(y)
    assert y.shape == y_hat.shape


def test_lstmrnn_sar():
    params = {
        "in_dim": 300,
        "hidden_dim": 8,
        "out_dim": 20,
        "num_layers": 2,
        "dropout": 0.1,
        "stream_sizes": [10, 10],
        "ar_orders": [2, 2],
        "init_type": "none",
    }
    model = LSTMRNNSAR(**params)
    assert model.prediction_type() == PredictionType.DETERMINISTIC
    _test_model_impl(model, params["in_dim"], params["out_dim"])

    y = torch.rand(8, 100, params["out_dim"])
    y_hat = model.preprocess_target(y)
    assert y.shape == y_hat.shape


def test_mdn():
    params = {
        "in_dim": 300,
        "hidden_dim": 8,
        "out_dim": 180,
        "num_layers": 2,
        "num_gaussians": 2,
        "dim_wise": True,
        "init_type": "none",
    }
    model = MDN(**params)
    assert model.prediction_type() == PredictionType.PROBABILISTIC
    _test_model_impl(model, params["in_dim"], params["out_dim"])


def test_mdnv2():
    params = {
        "in_dim": 300,
        "hidden_dim": 8,
        "out_dim": 180,
        "num_layers": 2,
        "dropout": 0.5,
        "num_gaussians": 2,
        "dim_wise": True,
        "init_type": "none",
    }
    model = MDNv2(**params)
    assert model.prediction_type() == PredictionType.PROBABILISTIC
    _test_model_impl(model, params["in_dim"], params["out_dim"])


def test_rmdn():
    params = {
        "in_dim": 300,
        "hidden_dim": 8,
        "out_dim": 180,
        "num_layers": 2,
        "num_gaussians": 2,
        "dim_wise": True,
        "init_type": "none",
    }
    model = RMDN(**params)
    assert model.prediction_type() == PredictionType.PROBABILISTIC
    _test_model_impl(model, params["in_dim"], params["out_dim"])


def test_ff_conv_lstm():
    params = {
        "in_dim": 300,
        "ff_hidden_dim": 8,
        "conv_hidden_dim": 8,
        "lstm_hidden_dim": 8,
        "dropout": 0.1,
        "num_lstm_layers": 2,
        "bidirectional": True,
        "out_dim": 180,
        "init_type": "none",
    }
    model = FFConvLSTM(**params)
    assert model.prediction_type() == PredictionType.DETERMINISTIC
    _test_model_impl(model, params["in_dim"], params["out_dim"])


def test_variance_predictor():
    params = {
        "in_dim": 300,
        "out_dim": 180,
        "num_layers": 2,
        "hidden_dim": 8,
        "kernel_size": 5,
        "dropout": 0.5,
        "init_type": "none",
    }
    model = VariancePredictor(**{**params, "use_mdn": False})
    assert model.prediction_type() == PredictionType.DETERMINISTIC
    _test_model_impl(model, params["in_dim"], params["out_dim"])

    model = VariancePredictor(**{**params, "use_mdn": True})
    assert model.prediction_type() == PredictionType.PROBABILISTIC
    _test_model_impl(model, params["in_dim"], params["out_dim"])



# -*- coding: utf-8 -*-

# Copyright 2022 Reo Yoneyama (Nagoya University)
#  MIT License (https://opensource.org/licenses/MIT)

"""Discriminator modules.

References:
    - https://github.com/bigpon/QPPWG
    - https://github.com/jik876/hifi-gan

"""

import copy
from logging import getLogger
from tkinter import W

# A logger for this file
logger = getLogger(__name__)

import torch
import torch.nn as nn
import torch.nn.functional as F
from nnsvs.usfgan.layers import Conv1d, Conv2d
from torchaudio.functional import spectrogram


class PWGDiscriminator(nn.Module):
    """Parallel WaveGAN Discriminator module."""

    def __init__(
        self,
        in_channels=1,
        out_channels=1,
        kernel_size=3,
        layers=10,
        conv_channels=64,
        dilation_factor=1,
        nonlinear_activation="LeakyReLU",
        nonlinear_activation_params={"negative_slope": 0.2},
        bias=True,
        use_weight_norm=True,
    ):
        """Initialize Parallel WaveGAN Discriminator module.

        Args:
            in_channels (int): Number of input channels.
            out_channels (int): Number of output channels.
            kernel_size (int): Number of output channels.
            layers (int): Number of conv layers.
            conv_channels (int): Number of chnn layers.
            dilation_factor (int): Dilation factor. For example, if dilation_factor = 2,
                the dilation will be 2, 4, 8, ..., and so on.
            nonlinear_activation (str): Nonlinear function after each conv.
            nonlinear_activation_params (dict): Nonlinear function parameters
            bias (bool): Whether to use bias parameter in conv.
            use_weight_norm (bool) Whether to use weight norm.
                If set to true, it will be applied to all of the conv layers.

        """
        super(PWGDiscriminator, self).__init__()
        assert (kernel_size - 1) % 2 == 0, "Not support even number kernel size."
        assert dilation_factor > 0, "Dilation factor must be > 0."
        self.conv_layers = nn.ModuleList()
        conv_in_channels = in_channels
        for i in range(layers - 1):
            if i == 0:
                dilation = 1
            else:
                dilation = i if dilation_factor == 1 else dilation_factor ** i
                conv_in_channels = conv_channels
            padding = (kernel_size - 1) // 2 * dilation
            conv_layer = [
                Conv1d(
                    conv_in_channels,
                    conv_channels,
                    kernel_size=kernel_size,
                    padding=padding,
                    dilation=dilation,
                    bias=bias,
                ),
                getattr(nn, nonlinear_activation)(
                    inplace=True, **nonlinear_activation_params
                ),
            ]
            self.conv_layers += conv_layer
        padding = (kernel_size - 1) // 2
        conv_last_layer = Conv1d(
            conv_in_channels,
            out_channels,
            kernel_size=kernel_size,
            padding=padding,
            bias=bias,
        )
        self.conv_layers += [conv_last_layer]

        # apply weight norm
        if use_weight_norm:
            self.apply_weight_norm()

    def forward(self, x, return_fmaps=False):
        """Calculate forward propagation.

        Args:
            x (Tensor): Input noise signal (B, 1, T).
            return_fmaps (bool): Whether to return feature maps.

        Returns:
            Tensor: Output tensor (B, 1, T)

        """
        fmaps = []
        for f in self.conv_layers:
            x = f(x)
            if return_fmaps:
                fmaps.append(x)

        if return_fmaps:
            return [x], fmaps
        else:
            return [x]

    def apply_weight_norm(self):
        """Apply weight normalization module from all of the layers."""

        def _apply_weight_norm(m):
            if isinstance(m, nn.Conv1d) or isinstance(m, nn.Conv2d):
                nn.utils.weight_norm(m)
                logger.debug(f"Weight norm is applied to {m}.")

        self.apply(_apply_weight_norm)

    def remove_weight_norm(self):
        """Remove weight normalization module from all of the layers."""

        def _remove_weight_norm(m):
            try:
                logger.debug(f"Weight norm is removed from {m}.")
                nn.utils.remove_weight_norm(m)
            except ValueError:  # this module didn't have weight norm
                return

        self.apply(_remove_weight_norm)


class HiFiGANPeriodDiscriminator(nn.Module):
    """HiFiGAN period discriminator module."""

    def __init__(
        self,
        in_channels=1,
        out_channels=1,
        period=3,
        kernel_sizes=[5, 3],
        channels=32,
        downsample_scales=[3, 3, 3, 3, 1],
        max_downsample_channels=1024,
        bias=True,
        nonlinear_activation="LeakyReLU",
        nonlinear_activation_params={"negative_slope": 0.1},
        use_weight_norm=True,
        use_spectral_norm=False,
    ):
        """Initialize HiFiGANPeriodDiscriminator module.

        Args:
            in_channels (int): Number of input channels.
            out_channels (int): Number of output channels.
            period (int): Period.
            kernel_sizes (list): Kernel sizes of initial conv layers and the final conv layer.
            channels (int): Number of initial channels.
            downsample_scales (list): List of downsampling scales.
            max_downsample_channels (int): Number of maximum downsampling channels.
            use_additional_convs (bool): Whether to use additional conv layers in residual blocks.
            bias (bool): Whether to add bias parameter in convolution layers.
            nonlinear_activation (str): Activation function module name.
            nonlinear_activation_params (dict): Hyperparameters for activation function.
            use_weight_norm (bool): Whether to use weight norm.
                If set to true, it will be applied to all of the conv layers.
            use_spectral_norm (bool): Whether to use spectral norm.
                If set to true, it will be applied to all of the conv layers.

        """
        super().__init__()
        assert len(kernel_sizes) == 2
        assert kernel_sizes[0] % 2 == 1, "Kernel size must be odd number."
        assert kernel_sizes[1] % 2 == 1, "Kernel size must be odd number."

        self.period = period
        self.convs = nn.ModuleList()
        in_chs = in_channels
        out_chs = channels
        for downsample_scale in downsample_scales:
            self.convs += [
                nn.Sequential(
                    nn.Conv2d(
                        in_chs,
                        out_chs,
                        (kernel_sizes[0], 1),
                        (downsample_scale, 1),
                        padding=((kernel_sizes[0] - 1) // 2, 0),
                    ),
                    getattr(nn, nonlinear_activation)(**nonlinear_activation_params),
                )
            ]
            in_chs = out_chs
            # NOTE(kan-bayashi): Use downsample_scale + 1?
            out_chs = min(out_chs * 4, max_downsample_channels)
        self.output_conv = nn.Conv2d(
            out_chs,
            out_channels,
            (kernel_sizes[1] - 1, 1),
            1,
            padding=((kernel_sizes[1] - 1) // 2, 0),
        )

        if use_weight_norm and use_spectral_norm:
            raise ValueError("Either use use_weight_norm or use_spectral_norm.")

        # apply weight norm
        if use_weight_norm:
            self.apply_weight_norm()

        # apply spectral norm
        if use_spectral_norm:
            self.apply_spectral_norm()

    def forward(self, x, return_fmaps=False):
        """Calculate forward propagation.

        Args:
            c (Tensor): Input tensor (B, in_channels, T).
            return_fmaps (bool): Whether to return feature maps.

        Returns:
            list: List of each layer's tensors.

        """
        # transform 1d to 2d -> (B, C, T/P, P)
        b, c, t = x.shape
        if t % self.period != 0:
            n_pad = self.period - (t % self.period)
            x = F.pad(x, (0, n_pad), "reflect")
            t += n_pad
        x = x.view(b, c, t // self.period, self.period)

        # forward conv
        fmap = []
        for f in self.convs:
            x = f(x)
            if return_fmaps:
                fmap.append(x)
        x = self.output_conv(x)
        out = torch.flatten(x, 1, -1)

        if return_fmaps:
            return out, fmap
        else:
            return out

    def apply_weight_norm(self):
        """Apply weight normalization module from all of the layers."""

        def _apply_weight_norm(m):
            if isinstance(m, nn.Conv2d):
                nn.utils.weight_norm(m)
                logger.debug(f"Weight norm is applied to {m}.")

        self.apply(_apply_weight_norm)

    def apply_spectral_norm(self):
        """Apply spectral normalization module from all of the layers."""

        def _apply_spectral_norm(m):
            if isinstance(m, nn.Conv2d):
                nn.utils.spectral_norm(m)
                logger.debug(f"Spectral norm is applied to {m}.")

        self.apply(_apply_spectral_norm)


class HiFiGANMultiPeriodDiscriminator(nn.Module):
    """HiFiGAN multi-period discriminator module."""

    def __init__(
        self,
        periods=[2, 3, 5, 7, 11],
        discriminator_params={
            "in_channels": 1,
            "out_channels": 1,
            "kernel_sizes": [5, 3],
            "channels": 32,
            "downsample_scales": [3, 3, 3, 3, 1],
            "max_downsample_channels": 1024,
            "bias": True,
            "nonlinear_activation": "LeakyReLU",
            "nonlinear_activation_params": {"negative_slope": 0.1},
            "use_weight_norm": True,
            "use_spectral_norm": False,
        },
    ):
        """Initialize HiFiGANMultiPeriodDiscriminator module.

        Args:
            periods (list): List of periods.
            discriminator_params (dict): Parameters for hifi-gan period discriminator module.
                The period parameter will be overwritten.

        """
        super().__init__()
        self.discriminators = nn.ModuleList()
        for period in periods:
            params = copy.deepcopy(discriminator_params)
            params["period"] = period
            self.discriminators += [HiFiGANPeriodDiscriminator(**params)]

    def forward(self, x, return_fmaps=False):
        """Calculate forward propagation.

        Args:
            x (Tensor): Input noise signal (B, 1, T).
            return_fmaps (bool): Whether to return feature maps.

        Returns:
            List: List of list of each discriminator outputs, which consists of each layer output tensors.

        """
        outs, fmaps = [], []
        for f in self.discriminators:
            if return_fmaps:
                out, fmap = f(x, return_fmaps)
                fmaps.extend(fmap)
            else:
                out = f(x)
            outs.append(out)

        if return_fmaps:
            return outs, fmaps
        else:
            return outs


class HiFiGANScaleDiscriminator(nn.Module):
    """HiFi-GAN scale discriminator module."""

    def __init__(
        self,
        in_channels=1,
        out_channels=1,
        kernel_sizes=[15, 41, 5, 3],
        channels=128,
        max_downsample_channels=1024,
        max_groups=16,
        bias=True,
        downsample_scales=[2, 2, 4, 4, 1],
        nonlinear_activation="LeakyReLU",
        nonlinear_activation_params={"negative_slope": 0.1},
        use_weight_norm=True,
        use_spectral_norm=False,
    ):
        """Initialize HiFiGAN scale discriminator module.

        Args:
            in_channels (int): Number of input channels.
            out_channels (int): Number of output channels.
            kernel_sizes (list): List of four kernel sizes. The first will be used for the first conv layer,
                and the second is for downsampling part, and the remaining two are for output layers.
            channels (int): Initial number of channels for conv layer.
            max_downsample_channels (int): Maximum number of channels for downsampling layers.
            bias (bool): Whether to add bias parameter in convolution layers.
            downsample_scales (list): List of downsampling scales.
            nonlinear_activation (str): Activation function module name.
            nonlinear_activation_params (dict): Hyperparameters for activation function.
            use_weight_norm (bool): Whether to use weight norm.
                If set to true, it will be applied to all of the conv layers.
            use_spectral_norm (bool): Whether to use spectral norm.
                If set to true, it will be applied to all of the conv layers.

        """
        super().__init__()
        self.layers = nn.ModuleList()

        # check kernel size is valid
        assert len(kernel_sizes) == 4
        for ks in kernel_sizes:
            assert ks % 2 == 1

        # add first layer
        self.layers += [
            nn.Sequential(
                nn.Conv1d(
                    in_channels,
                    channels,
                    # NOTE(kan-bayashi): Use always the same kernel size
                    kernel_sizes[0],
                    bias=bias,
                    padding=(kernel_sizes[0] - 1) // 2,
                ),
                getattr(nn, nonlinear_activation)(**nonlinear_activation_params),
            )
        ]

        # add downsample layers
        in_chs = channels
        out_chs = channels
        # NOTE(kan-bayashi): Remove hard coding?
        groups = 4
        for downsample_scale in downsample_scales:
            self.layers += [
                nn.Sequential(
                    nn.Conv1d(
                        in_chs,
                        out_chs,
                        kernel_size=kernel_sizes[1],
                        stride=downsample_scale,
                        padding=(kernel_sizes[1] - 1) // 2,
                        groups=groups,
                        bias=bias,
                    ),
                    getattr(nn, nonlinear_activation)(**nonlinear_activation_params),
                )
            ]
            in_chs = out_chs
            # NOTE(kan-bayashi): Remove hard coding?
            out_chs = min(in_chs * 2, max_downsample_channels)
            # NOTE(kan-bayashi): Remove hard coding?
            groups = min(groups * 4, max_groups)

        # add final layers
        out_chs = min(in_chs * 2, max_downsample_channels)
        self.layers += [
            nn.Sequential(
                nn.Conv1d(
                    in_chs,
                    out_chs,
                    kernel_size=kernel_sizes[2],
                    stride=1,
                    padding=(kernel_sizes[2] - 1) // 2,
                    bias=bias,
                ),
                getattr(nn, nonlinear_activation)(**nonlinear_activation_params),
            )
        ]
        self.last_layer = nn.Conv1d(
            out_chs,
            out_channels,
            kernel_size=kernel_sizes[3],
            stride=1,
            padding=(kernel_sizes[3] - 1) // 2,
            bias=bias,
        )

        if use_weight_norm and use_spectral_norm:
            raise ValueError("Either use use_weight_norm or use_spectral_norm.")

        # apply weight norm
        if use_weight_norm:
            self.apply_weight_norm()

        # apply spectral norm
        if use_spectral_norm:
            self.apply_spectral_norm()

    def forward(self, x, return_fmaps=False):
        """Calculate forward propagation.

        Args:
            x (Tensor): Input noise signal (B, 1, T).
            return_fmaps (bool): Whether to return feature maps.

        Returns:
            List: List of output tensors of each layer.

        """
        fmap = []
        for f in self.layers:
            x = f(x)
            if return_fmaps:
                fmap.append(x)
        out = self.last_layer(x)

        if return_fmaps:
            return out, fmap
        else:
            return out

    def apply_weight_norm(self):
        """Apply weight normalization module from all of the layers."""

        def _apply_weight_norm(m):
            if isinstance(m, nn.Conv2d):
                nn.utils.weight_norm(m)
                logger.debug(f"Weight norm is applied to {m}.")

        self.apply(_apply_weight_norm)

    def apply_spectral_norm(self):
        """Apply spectral normalization module from all of the layers."""

        def _apply_spectral_norm(m):
            if isinstance(m, nn.Conv2d):
                nn.utils.spectral_norm(m)
                logger.debug(f"Spectral norm is applied to {m}.")

        self.apply(_apply_spectral_norm)


class HiFiGANMultiScaleDiscriminator(nn.Module):
    """HiFi-GAN multi-scale discriminator module."""

    def __init__(
        self,
        scales=3,
        downsample_pooling="AvgPool1d",
        # follow the official implementation setting
        downsample_pooling_params={
            "kernel_size": 4,
            "stride": 2,
            "padding": 2,
        },
        discriminator_params={
            "in_channels": 1,
            "out_channels": 1,
            "kernel_sizes": [15, 41, 5, 3],
            "channels": 128,
            "max_downsample_channels": 1024,
            "max_groups": 16,
            "bias": True,
            "downsample_scales": [2, 2, 4, 4, 1],
            "nonlinear_activation": "LeakyReLU",
            "nonlinear_activation_params": {"negative_slope": 0.1},
        },
        follow_official_norm=False,
    ):
        """Initialize HiFiGAN multi-scale discriminator module.

        Args:
            scales (int): Number of multi-scales.
            downsample_pooling (str): Pooling module name for downsampling of the inputs.
            downsample_pooling_params (dict): Parameters for the above pooling module.
            discriminator_params (dict): Parameters for hifi-gan scale discriminator module.
            follow_official_norm (bool): Whether to follow the norm setting of the official
                implementation. The first discriminator uses spectral norm and the other
                discriminators use weight norm.

        """
        super().__init__()
        self.discriminators = nn.ModuleList()

        # add discriminators
        for i in range(scales):
            params = copy.deepcopy(discriminator_params)
            if follow_official_norm:
                if i == 0:
                    params["use_weight_norm"] = False
                    params["use_spectral_norm"] = True
                else:
                    params["use_weight_norm"] = True
                    params["use_spectral_norm"] = False
            self.discriminators += [HiFiGANScaleDiscriminator(**params)]
        self.pooling = getattr(nn, downsample_pooling)(**downsample_pooling_params)

    def forward(self, x, return_fmaps=False):
        """Calculate forward propagation.

        Args:
            x (Tensor): Input noise signal (B, 1, T).
            return_fmaps (bool): Whether to return feature maps.

        Returns:
            List: List of list of each discriminator outputs, which consists of each layer output tensors.

        """
        outs, fmaps = [], []
        for f in self.discriminators:
            if return_fmaps:
                out, fmap = f(x, return_fmaps)
                fmaps.extend(fmap)
            else:
                out = f(x)
            outs.append(out)
            x = self.pooling(x)

        if return_fmaps:
            return outs, fmaps
        else:
            return outs


class HiFiGANMultiScaleMultiPeriodDiscriminator(nn.Module):
    """HiFi-GAN multi-scale + multi-period discriminator module."""

    def __init__(
        self,
        # Multi-scale discriminator related
        scales=3,
        scale_downsample_pooling="AvgPool1d",
        scale_downsample_pooling_params={
            "kernel_size": 4,
            "stride": 2,
            "padding": 2,
        },
        scale_discriminator_params={
            "in_channels": 1,
            "out_channels": 1,
            "kernel_sizes": [15, 41, 5, 3],
            "channels": 128,
            "max_downsample_channels": 1024,
            "max_groups": 16,
            "bias": True,
            "downsample_scales": [2, 2, 4, 4, 1],
            "nonlinear_activation": "LeakyReLU",
            "nonlinear_activation_params": {"negative_slope": 0.1},
        },
        follow_official_norm=True,
        # Multi-period discriminator related
        periods=[2, 3, 5, 7, 11],
        period_discriminator_params={
            "in_channels": 1,
            "out_channels": 1,
            "kernel_sizes": [5, 3],
            "channels": 32,
            "downsample_scales": [3, 3, 3, 3, 1],
            "max_downsample_channels": 1024,
            "bias": True,
            "nonlinear_activation": "LeakyReLU",
            "nonlinear_activation_params": {"negative_slope": 0.1},
            "use_weight_norm": True,
            "use_spectral_norm": False,
        },
    ):
        """Initialize HiFiGAN multi-scale + multi-period discriminator module.

        Args:
            scales (int): Number of multi-scales.
            scale_downsample_pooling (str): Pooling module name for downsampling of the inputs.
            scale_downsample_pooling_params (dict): Parameters for the above pooling module.
            scale_discriminator_params (dict): Parameters for hifi-gan scale discriminator module.
            follow_official_norm (bool): Whether to follow the norm setting of the official
                implementation. The first discriminator uses spectral norm and the other
                discriminators use weight norm.
            periods (list): List of periods.
            period_discriminator_params (dict): Parameters for hifi-gan period discriminator module.
                The period parameter will be overwritten.

        """
        super().__init__()
        self.msd = HiFiGANMultiScaleDiscriminator(
            scales=scales,
            downsample_pooling=scale_downsample_pooling,
            downsample_pooling_params=scale_downsample_pooling_params,
            discriminator_params=scale_discriminator_params,
            follow_official_norm=follow_official_norm,
        )
        self.mpd = HiFiGANMultiPeriodDiscriminator(
            periods=periods,
            discriminator_params=period_discriminator_params,
        )

    def forward(self, x, return_fmaps=False):
        """Calculate forward propagation.

        Args:
            x (Tensor): Input noise signal (B, 1, T).
            return_fmaps (bool): Whether to return feature maps.

        Returns:
            List: List of list of each discriminator outputs,
                which consists of each layer output tensors.
                Multi scale and multi period ones are concatenated.

        """
        if return_fmaps:
            msd_outs, msd_fmaps = self.msd(x, return_fmaps)
            mpd_outs, mpd_fmaps = self.mpd(x, return_fmaps)
            outs = msd_outs + mpd_outs
            fmaps = msd_fmaps + mpd_fmaps
            return outs, fmaps
        else:
            msd_outs = self.msd(x)
            mpd_outs = self.mpd(x)
            outs = msd_outs + mpd_outs
            return outs


class UnivNetSpectralDiscriminator(nn.Module):
    """UnivNet spectral discriminator module."""

    def __init__(
        self,
        fft_size,
        hop_size,
        win_length,
        window="hann_window",
        kernel_sizes=[(3, 9), (3, 9), (3, 9), (3, 9), (3, 3), (3, 3)],
        strides=[(1, 1), (1, 2), (1, 2), (1, 2), (1, 1), (1, 1)],
        channels=32,
        bias=True,
        nonlinear_activation="LeakyReLU",
        nonlinear_activation_params={"negative_slope": 0.2},
        use_weight_norm=True,
    ):
        """Initialize HiFiGAN scale discriminator module.

        Args:
            kernel_sizes (list): List of four kernel sizes. The first will be used for the first conv layer,
                and the second is for downsampling part, and the remaining two are for output layers.
            strides (list):
            channels (int): Initial number of channels for conv layer.
            bias (bool): Whether to add bias parameter in convolution layers.
            nonlinear_activation (str): Activation function module name.
            nonlinear_activation_params (dict): Hyperparameters for activation function.
            use_weight_norm (bool): Whether to use weight norm.
                If set to true, it will be applied to all of the conv layers.

        """
        super().__init__()

        self.fft_size = fft_size
        self.hop_size = hop_size
        self.win_length = win_length
        self.register_buffer("window", getattr(torch, window)(win_length))

        self.layers = nn.ModuleList()

        # check kernel size is valid
        assert len(kernel_sizes) == len(strides)

        # add first layer
        self.layers += [
            nn.Sequential(
                nn.Conv2d(
                    1,
                    channels,
                    kernel_sizes[0],
                    stride=strides[0],
                    bias=bias,
                ),
                getattr(nn, nonlinear_activation)(**nonlinear_activation_params),
            )
        ]

        for i in range(1, len(kernel_sizes) - 2):
            self.layers += [
                nn.Sequential(
                    nn.Conv2d(
                        channels,
                        channels,
                        kernel_size=kernel_sizes[i],
                        stride=strides[i],
                        bias=bias,
                    ),
                    getattr(nn, nonlinear_activation)(**nonlinear_activation_params),
                )
            ]

        # add final layers
        self.layers += [
            nn.Sequential(
                nn.Conv2d(
                    channels,
                    channels,
                    kernel_size=kernel_sizes[-2],
                    stride=strides[-2],
                    bias=bias,
                ),
                getattr(nn, nonlinear_activation)(**nonlinear_activation_params),
            )
        ]
        self.layers += [
            nn.Conv2d(
                channels,
                1,
                kernel_size=kernel_sizes[-1],
                stride=strides[-1],
                bias=bias,
            )
        ]

        # apply weight norm
        if use_weight_norm:
            self.apply_weight_norm()

    def forward(self, x, return_fmaps=False):
        """Calculate forward propagation.

        Args:
            x (Tensor): Input noise signal (B, 1, T).
            return_fmaps (bool): Whether to return feature maps.

        Returns:
            List: List of output tensors of each layer.

        """
        x = spectrogram(
            x,
            pad=self.win_length // 2,
            window=self.window,
            n_fft=self.fft_size,
            hop_length=self.hop_size,
            win_length=self.win_length,
            power=1.0,
            normalized=False,
        ).transpose(-1, -2)

        fmap = []
        for f in self.layers:
            x = f(x)
            if return_fmaps:
                fmap.append(x)

        if return_fmaps:
            return x, fmap
        else:
            return x

    def apply_weight_norm(self):
        """Apply weight normalization module from all of the layers."""

        def _apply_weight_norm(m):
            if isinstance(m, nn.Conv2d):
                nn.utils.weight_norm(m)
                logger.debug(f"Weight norm is applied to {m}.")

        self.apply(_apply_weight_norm)


class UnivNetMultiResolutionSpectralDiscriminator(nn.Module):
    """UnivNet multi-resolution spectral discriminator module."""

    def __init__(
        self,
        fft_sizes=[1024, 2048, 512],
        hop_sizes=[120, 240, 50],
        win_lengths=[600, 1200, 240],
        window="hann_window",
        discriminator_params={
            "channels": 32,
            "kernel_sizes": [(3, 9), (3, 9), (3, 9), (3, 9), (3, 3), (3, 3)],
            "strides": [(1, 1), (1, 2), (1, 2), (1, 2), (1, 1), (1, 1)],
            "bias": True,
            "nonlinear_activation": "LeakyReLU",
            "nonlinear_activation_params": {"negative_slope": 0.2},
        },
    ):
        """Initialize UnivNetMultiResolutionSpectralDiscriminator module.

        Args:
            scales (int): Number of multi-scales.
            downsample_pooling (str): Pooling module name for downsampling of the inputs.
            downsample_pooling_params (dict): Parameters for the above pooling module.
            discriminator_params (dict): Parameters for hifi-gan scale discriminator module.
            follow_official_norm (bool): Whether to follow the norm setting of the official
                implementation. The first discriminator uses spectral norm and the other
                discriminators use weight norm.

        """
        super().__init__()
        assert len(fft_sizes) == len(hop_sizes) == len(win_lengths)
        self.discriminators = nn.ModuleList()

        # add discriminators
        for i in range(len(fft_sizes)):
            params = copy.deepcopy(discriminator_params)
            self.discriminators += [
                UnivNetSpectralDiscriminator(
                    fft_size=fft_sizes[i],
                    hop_size=hop_sizes[i],
                    win_length=win_lengths[i],
                    window=window,
                    **params,
                )
            ]

    def forward(self, x, return_fmaps=False):
        """Calculate forward propagation.

        Args:
            x (Tensor): Input noise signal (B, 1, T).
            return_fmaps (bool): Whether to return feature maps.

        Returns:
            List: List of list of each discriminator outputs, which consists of each layer output tensors.

        """
        outs, fmaps = [], []
        for f in self.discriminators:
            if return_fmaps:
                out, fmap = f(x, return_fmaps)
                fmaps.extend(fmap)
            else:
                out = f(x)
            outs.append(out)

        if return_fmaps:
            return outs, fmaps
        else:
            return outs


class UnivNetMultiResolutionMultiPeriodDiscriminator(nn.Module):
    """UnivNet multi-resolution + multi-period discriminator module."""

    def __init__(
        self,
        # Multi-resolution discriminator related
        fft_sizes=[1024, 2048, 512],
        hop_sizes=[120, 240, 50],
        win_lengths=[600, 1200, 240],
        window="hann_window",
        spectral_discriminator_params={
            "channels": 32,
            "kernel_sizes": [(3, 9), (3, 9), (3, 9), (3, 9), (3, 3), (3, 3)],
            "strides": [(1, 1), (1, 2), (1, 2), (1, 2), (1, 1), (1, 1)],
            "bias": True,
            "nonlinear_activation": "LeakyReLU",
            "nonlinear_activation_params": {"negative_slope": 0.2},
        },
        # Multi-period discriminator related
        periods=[2, 3, 5, 7, 11],
        period_discriminator_params={
            "in_channels": 1,
            "out_channels": 1,
            "kernel_sizes": [5, 3],
            "channels": 32,
            "downsample_scales": [3, 3, 3, 3, 1],
            "max_downsample_channels": 1024,
            "bias": True,
            "nonlinear_activation": "LeakyReLU",
            "nonlinear_activation_params": {"negative_slope": 0.1},
            "use_weight_norm": True,
            "use_spectral_norm": False,
        },
    ):
        """Initialize UnivNetMultiResolutionMultiPeriodDiscriminator module.

        Args:
            sperctral_discriminator_params (dict): Parameters for hifi-gan scale discriminator module.
            periods (list): List of periods.
            period_discriminator_params (dict): Parameters for hifi-gan period discriminator module.
                The period parameter will be overwritten.

        """
        super().__init__()
        self.mrd = UnivNetMultiResolutionSpectralDiscriminator(
            fft_sizes=fft_sizes,
            hop_sizes=hop_sizes,
            win_lengths=win_lengths,
            window=window,
            discriminator_params=spectral_discriminator_params,
        )
        self.mpd = HiFiGANMultiPeriodDiscriminator(
            periods=periods,
            discriminator_params=period_discriminator_params,
        )

    def forward(self, x, return_fmaps=False):
        """Calculate forward propagation.

        Args:
            x (Tensor): Input noise signal (B, 1, T).
            return_fmaps (bool): Whether to return feature maps.

        Returns:
            List: List of list of each discriminator outputs,
                which consists of each layer output tensors.
                Multi scale and multi period ones are concatenated.

        """
        if return_fmaps:
            mrd_outs, mrd_fmaps = self.mrd(x, return_fmaps)
            mpd_outs, mpd_fmaps = self.mpd(x, return_fmaps)
            outs = mrd_outs + mpd_outs
            fmaps = mrd_fmaps + mpd_fmaps
            return outs, fmaps
        else:
            mrd_outs = self.mrd(x)
            mpd_outs = self.mpd(x)
            outs = mrd_outs + mpd_outs
            return outs



# NOTE: Opencpop uses SP as silence and AP as aspire (breath)
# for consistency with the other JP database, adding sil, pau and br to the list.
phonemes = [
    "AP",
    "SP",
    "sil",
    "pau",
    "br",
    "a",
    "ai",
    "an",
    "ang",
    "ao",
    "b",
    "c",
    "ch",
    "d",
    "e",
    "ei",
    "en",
    "eng",
    "er",
    "f",
    "g",
    "h",
    "i",
    "ia",
    "ian",
    "iang",
    "iao",
    "ie",
    "in",
    "ing",
    "iong",
    "iu",
    "j",
    "k",
    "l",
    "m",
    "n",
    "o",
    "ong",
    "ou",
    "p",
    "q",
    "r",
    "s",
    "sh",
    "t",
    "u",
    "ua",
    "uai",
    "uan",
    "uang",
    "ui",
    "un",
    "uo",
    "v",
    "van",
    "ve",
    "vn",
    "w",
    "x",
    "y",
    "z",
    "zh",
]


_pad = "~"

symbols = [_pad] + phonemes


_symbol_to_id = {s: i for i, s in enumerate(symbols)}
_id_to_symbol = {i: s for i, s in enumerate(symbols)}


def num_vocab():
    return len(symbols)


def text_to_sequence(text):
    return [_symbol_to_id[s] for s in text]


def sequence_to_text(seq):
    return [_id_to_symbol[s] for s in seq]



phonemes = [
    "A",
    "E",
    "I",
    "N",
    "O",
    "U",
    "a",
    "b",
    "br",
    "by",
    "ch",
    "cl",
    "d",
    "dy",
    "e",
    "f",
    "g",
    "gy",
    "h",
    "hy",
    "i",
    "j",
    "k",
    "ky",
    "m",
    "my",
    "n",
    "ny",
    "o",
    "p",
    "py",
    "r",
    "ry",
    "s",
    "sh",
    "t",
    "ts",
    "ty",
    "u",
    "v",
    "w",
    "y",
    "z",
    "pau",
    "sil",
    "fy",
    "vy",
    "GlottalStop",
    "Edge",
]

_pad = "~"

symbols = [_pad] + phonemes


_symbol_to_id = {s: i for i, s in enumerate(symbols)}
_id_to_symbol = {i: s for i, s in enumerate(symbols)}


def num_vocab():
    return len(symbols)


def text_to_sequence(text):
    return [_symbol_to_id[s] for s in text]


def sequence_to_text(seq):
    return [_id_to_symbol[s] for s in seq]



import torch
from nnsvs.wavenet.modules import Conv1d1x1, ResSkipBlock
from torch import nn
from torch.nn import functional as F


class WaveNet(nn.Module):
    """WaveNet

    Args:
        in_dim (int): the dimension of the input
        out_dim (int): the dimension of the output
        layers (int): the number of layers
        stacks (int): the number of residual stacks
        residual_channels (int): the number of residual channels
        gate_channels (int): the number of channels for the gating function
        skip_out_channels (int): the number of channels in the skip output
        kernel_size (int): the size of the convolutional kernel
    """

    def __init__(
        self,
        in_dim=334,
        out_dim=206,
        layers=10,
        stacks=1,
        residual_channels=64,
        gate_channels=128,
        skip_out_channels=64,
        kernel_size=3,
    ):
        super().__init__()
        self.in_dim = in_dim
        self.out_dim = out_dim
        self.first_conv = Conv1d1x1(out_dim, residual_channels)

        self.main_conv_layers = nn.ModuleList()
        layers_per_stack = layers // stacks
        for layer in range(layers):
            dilation = 2 ** (layer % layers_per_stack)
            conv = ResSkipBlock(
                residual_channels,
                gate_channels,
                kernel_size,
                skip_out_channels,
                dilation=dilation,
                cin_channels=in_dim,
            )
            self.main_conv_layers.append(conv)

        self.last_conv_layers = nn.ModuleList(
            [
                nn.ReLU(),
                Conv1d1x1(skip_out_channels, skip_out_channels),
                nn.ReLU(),
                Conv1d1x1(skip_out_channels, out_dim),
            ]
        )

    def forward(self, c, x, lengths=None):
        """Forward step

        Args:
            c (torch.Tensor): the conditional features (B, T, C)
            x (torch.Tensor): the target features (B, T, C)

        Returns:
            torch.Tensor: the output waveform
        """
        x = x.transpose(1, 2)
        c = c.transpose(1, 2)

        x = self.first_conv(x)

        skips = 0
        for f in self.main_conv_layers:
            x, h = f(x, c)
            skips += h

        x = skips
        for f in self.last_conv_layers:
            x = f(x)

        # (B, C, T) -> (B, T, C)
        x = x.transpose(1, 2)

        return x

    def inference(self, c, num_time_steps=100, tqdm=lambda x: x):
        """Inference step

        Args:
            c (torch.Tensor): the local conditioning feature (B, T, C)
            num_time_steps (int): the number of time steps to generate
            tqdm (lambda): a tqdm function to track progress

        Returns:
            torch.Tensor: the output waveform
        """
        self.clear_buffer()

        # Local conditioning
        B = c.shape[0]

        outputs = []

        # 自己回帰生成における初期値
        current_input = torch.zeros(B, 1, self.out_dim).to(c.device)

        if tqdm is None:
            ts = range(num_time_steps)
        else:
            ts = tqdm(range(num_time_steps))

        # 逐次的に生成
        for t in ts:
            # 時刻 t における入力は、時刻 t-1 における出力
            if t > 0:
                current_input = outputs[-1]

            # 時刻 t における条件付け特徴量
            ct = c[:, t, :].unsqueeze(1)

            x = current_input
            x = self.first_conv.incremental_forward(x)
            skips = 0
            for f in self.main_conv_layers:
                x, h = f.incremental_forward(x, ct)
                skips += h
            x = skips
            for f in self.last_conv_layers:
                if hasattr(f, "incremental_forward"):
                    x = f.incremental_forward(x)
                else:
                    x = f(x)
            # Softmax によって、出力をカテゴリカル分布のパラメータに変換
            x = F.softmax(x.view(B, -1), dim=1)
            # カテゴリカル分布からサンプリング
            x = torch.distributions.OneHotCategorical(x).sample()
            outputs += [x.data]

        # T x B x C
        # 各時刻における出力を結合
        outputs = torch.stack(outputs)
        # B x T x C
        outputs = outputs.transpose(0, 1).contiguous()

        self.clear_buffer()
        return outputs

    def clear_buffer(self):
        """Clear the internal buffer."""
        self.first_conv.clear_buffer()
        for f in self.main_conv_layers:
            f.clear_buffer()
        for f in self.last_conv_layers:
            try:
                f.clear_buffer()
            except AttributeError:
                pass

    def remove_weight_norm_(self):
        """Remove weight normalization of the model"""

        def _remove_weight_norm(m):
            try:
                torch.nn.utils.remove_weight_norm(m)
            except ValueError:
                return

        self.apply(_remove_weight_norm)



from nnsvs.util import get_world_stream_info


def test_get_world_stream_info():
    assert get_world_stream_info(24000, 59, 1, vibrato_mode="none") == [60, 1, 1, 3]
    assert get_world_stream_info(24000, 59, 1, vibrato_mode="diff") == [60, 1, 1, 3, 1]
    assert get_world_stream_info(24000, 59, 1, vibrato_mode="sine") == [
        60,
        1,
        1,
        3,
        3,
        1,
    ]

    assert get_world_stream_info(44100, 59, 1, vibrato_mode="none") == [60, 1, 1, 5]
    assert get_world_stream_info(48000, 59, 1, vibrato_mode="none") == [60, 1, 1, 5]


def test_get_world_stream_info_mcep_aperiodicity():
    assert (
        get_world_stream_info(
            44100,
            59,
            1,
            vibrato_mode="none",
            use_mcep_aperiodicity=True,
            mcep_aperiodicity_order=24,
        )
        == [60, 1, 1, 25]
    )
    assert (
        get_world_stream_info(
            48000,
            59,
            1,
            vibrato_mode="none",
            use_mcep_aperiodicity=True,
            mcep_aperiodicity_order=24,
        )
        == [60, 1, 1, 25]
    )
